import pytest
import numpy as np
import pandas as pd
import matplotlib as mpl
from matplotlib import patches
import matplotlib.pyplot as plt
import iplotx as ipx


def vertex_collection(**kwargs):
    if "layout" not in kwargs:
        kwargs["layout"] = pd.DataFrame(
            data=np.array([[0, 0], [1, 1]], np.float64),
            index=["A", "B"],
        )
    return ipx.artists.VertexCollection(**kwargs)


def edge_collection(**kwargs):
    if vertex_collection not in kwargs:
        kwargs["vertex_collection"] = vertex_collection()
    if "vertex_ids" not in kwargs:
        vids = kwargs["vertex_collection"].get_index()
        nedges = 4
        kwargs["vertex_ids"] = [
            [vids[i % len(vids)], vids[(i + 1) % len(vids)]] for i in range(nedges)
        ]
        kwargs["patches"] = [
            patches.PathPatch(mpl.path.Path(np.array([[0, 0], [1, 1]]))),
        ] * len(kwargs["vertex_ids"])

    return ipx.artists.EdgeCollection(**kwargs)


def test_arrows_init():
    edges = edge_collection()
    res = ipx.artists.EdgeArrowCollection(
        edges,
    )
    assert isinstance(res, ipx.artists.EdgeArrowCollection)


def test_cmap():
    edges = edge_collection(
        style={
            "cmap": mpl.colormaps.get_cmap("viridis"),
            "norm": mpl.colors.Normalize(vmin=0, vmax=1),
        }
    )
    res = ipx.artists.EdgeArrowCollection(
        edges,
    )
    assert isinstance(res, ipx.artists.EdgeArrowCollection)


def test_properties():
    res = ipx.artists.EdgeArrowCollection(
        edge_collection(),
    )
    np.testing.assert_array_almost_equal(res.get_sizes(), np.array([10.4] * 4))
    np.testing.assert_array_almost_equal(res.get_sizes_dpi(), np.array([10.4] * 4))
    res.set_sizes(None)

    fig = plt.figure()
    try:
        res.set_figure(fig)
        assert res.get_figure() == fig
    finally:
        plt.close(fig)

    with pytest.raises(ValueError):
        res.set_array(None)

    res = ipx.artists.EdgeArrowCollection(
        edge_collection(),
    )
    res.set_colors(np.asarray([(0, 0, 1, 1)] * 4))


def test_make_arrow_patch():
    markers = [
        "|>",
        "|/",
        "|\\",
        ">",
        ">>",
        ")>",
        ")",
        "|",
        "s",
        "d",
        "p",
        "q",
        "none",
    ]
    for marker in markers:
        patch, size = ipx.edge.arrow.make_arrow_patch(
            marker,
            width=10,
            height="width",
            color="blue",
        )
        assert isinstance(patch, patches.Patch)
        assert size == 10

    with pytest.raises(ValueError):
        ipx.edge.arrow.make_arrow_patch(
            "unsupported_marker",
        )



"""
Basics of styling
======================

This example shows a few simple styling options available in `iplotx`.

.. tip::
  `iplotx` has an internal library of styles. See the :doc:`plot_multistyle`
  example.
"""

import igraph as ig
import matplotlib.pyplot as plt
import iplotx as ipx

g = ig.Graph.Ring(5)
layout = g.layout("circle")
style = {
    "vertex": {
        "facecolor": ["purple", "pink", "black"],
        "edgecolor": "black",
    },
    "edge": {
        "linestyle": "--",
        "linewidth": 2,
        "color": "deeppink",
    },
}
fig, ax = plt.subplots(figsize=(3, 3))
ipx.network(g, ax=ax, layout=layout, style=style)

# %%
# You can also **chain styles**. For example, you can apply an existing
# style first, and then modify it with a custom dictionary:

style = {
    "edge": {
        "linestyle": "--",
    },
}
fig, ax = plt.subplots(figsize=(3, 3))
ipx.network(
    g,
    ax=ax,
    layout=layout,
    style=["rededge", style],
)

# %%
# As you can see, the network is visualised with the "rededge" style (which
# has red, solid edges) but further modified to have dashed lines.



"""
Network element showroom
========================

This example showcases various plot elements that can be styled in ``iplotx``.
"""

import matplotlib.pyplot as plt
import iplotx as ipx

network = {
    "edges": [(0, 1), (1, 2), (2, 3), (3, 0), (2, 2)],
    "directed": True,
}
layout = [(0, 0), (1, 0), (1, 1), (0, 1)]

fig, ax = plt.subplots(figsize=(4.3, 4.3))
ipx.network(
    network,
    layout,
    ax=ax,
    vertex_labels=["Vertex\nlabel"] + [""] * 3,
    edge_labels=["", "", "Edge\nlabel", "", ""],
    vertex_label_bbox_edgecolor="k",
    vertex_label_bbox_linewidth=1.5,
    vertex_label_bbox_facecolor="steelblue",
    vertex_label_bbox_alpha=0.5,
    edge_label_bbox_edgecolor="k",
    edge_label_bbox_linewidth=1.5,
    edge_label_bbox_facecolor="tomato",
    edge_color=["k", "tomato", "k", "k", "k"],
    edge_arrow_color=["k", "darkred", "k", "k", "k"],
    edge_linewidth=[2, 4, 2, 2, 2],
    vertex_facecolor=["w"] * 3 + ["steelblue"],
    vertex_edgecolor="k",
    vertex_linewidth=2,
    vertex_size=[50, 25, 25, 50],
    vertex_marker="o",
    vertex_label_color="k",
    edge_curved=True,
    edge_tension=[0, 1.25, 0, 3.5],
    margins=(0, 0.08),
)



import pytest
import numpy as np
import matplotlib as mpl
from iplotx.edge import geometry


def test_loops_per_angle_empty():
    res = geometry._compute_loops_per_angle(10, [])
    assert res == [(0, 2 * np.pi, 10)]


def test_shorter_edge_invalid():
    path = mpl.path.Path(np.array([[0, 0], [1, 1]]))
    with pytest.raises(ValueError):
        geometry._get_shorter_edge_coords(
            path,
            5,
            -0.5,
        )


def test_compute_path_straight_invalid():
    with pytest.raises(ValueError):
        geometry._compute_edge_path_straight(
            None,
            None,
            None,
            None,
            None,
            layout_coordinate_system="invalid",
        )


@pytest.fixture
def args_make_paths():
    vpath = mpl.path.Path(np.array([[-1, -1], [1, -1], [1, 1], [-1, 1], [-1, -1]]))
    transform = mpl.transforms.IdentityTransform()
    trans = transform.transform
    trans_inv = transform.inverted().transform
    return (vpath, trans, trans_inv)


def test_compute_path_straight_polar(args_make_paths):
    vpath, trans, trans_inv = args_make_paths
    epath, angles = geometry._compute_edge_path_straight(
        [[0, 0], [10, 0]],
        [vpath] * 2,
        (2, 2),
        trans,
        trans_inv,
        layout_coordinate_system="polar",
    )
    assert isinstance(epath, mpl.path.Path)
    assert len(angles) == 2
    assert isinstance(angles[0], float)
    assert isinstance(angles[1], float)
    assert np.round(angles[0] - angles[1] + np.pi, 2) == 0.0


@pytest.mark.parametrize("waypoints,expected", [["y0x1", [10, 0]], ["x0y1", [0, 5]], [[0, 5]] * 2])
def test_compute_waypoints(args_make_paths, waypoints, expected):
    vpath, trans, trans_inv = args_make_paths
    epath, angles = geometry._compute_edge_path_waypoints(
        waypoints,
        [[0, 0], [10, 5]],
        [vpath] * 2,
        (2, 2),
        trans,
        trans_inv,
    )
    assert isinstance(epath, mpl.path.Path)
    assert len(epath.vertices) == 3
    assert list(epath.vertices[1]) == expected


def test_compute_waypoints_port(args_make_paths):
    vpath, trans, trans_inv = args_make_paths
    epath, angles = geometry._compute_edge_path_waypoints(
        "x0y1",
        [[0, 0], [10, 5]],
        [vpath] * 2,
        (2, 2),
        trans,
        trans_inv,
        ports=("n", "w"),
    )
    assert isinstance(epath, mpl.path.Path)
    assert len(epath.vertices) == 3
    assert list(epath.vertices[1]) == [0, 5]


@pytest.mark.parametrize(
    "waypoints,expected",
    [
        ["xmidy0,xmidy1", [[5, 0], [5, 5]]],
        ["x0ymid,x1ymid", [[0, 2.5], [10, 2.5]]],
    ],
)
def test_double_waypoints(args_make_paths, waypoints, expected):
    vpath, trans, trans_inv = args_make_paths
    epath, angles = geometry._compute_edge_path_waypoints(
        waypoints,
        [[0, 0], [10, 5]],
        [vpath] * 2,
        (2, 2),
        trans,
        trans_inv,
    )
    assert isinstance(epath, mpl.path.Path)
    assert len(epath.vertices) == 4
    assert list(epath.vertices[1]) == expected[0]
    assert list(epath.vertices[2]) == expected[1]


@pytest.mark.parametrize(
    "waypoints,expected,port",
    [
        ["xmidy0,xmidy1", [[5, 0], [5, 5]], "w"],
        ["x0ymid,x1ymid", [[0, 2.5], [10, 2.5]], "s"],
    ],
)
def test_double_waypoints_ports(args_make_paths, waypoints, expected, port):
    vpath, trans, trans_inv = args_make_paths
    epath, angles = geometry._compute_edge_path_waypoints(
        waypoints,
        [[0, 0], [10, 5]],
        [vpath] * 2,
        (2, 2),
        trans,
        trans_inv,
        ports=(None, port),
    )
    assert isinstance(epath, mpl.path.Path)
    assert len(epath.vertices) == 4
    assert list(epath.vertices[1]) == expected[0]
    assert list(epath.vertices[2]) == expected[1]



from typing import (
    Optional,
    Sequence,
    Any,
)
from collections.abc import Hashable
from collections import defaultdict

import numpy as np
import pandas as pd
import matplotlib as mpl

from .style import (
    context,
    get_style,
    rotate_style,
    merge_styles,
)
from .utils.matplotlib import (
    _stale_wrapper,
    _forwarder,
    _build_cmap_fun,
)
from .ingest import (
    ingest_tree_data,
    data_providers,
)
from .vertex import (
    VertexCollection,
)
from .edge import (
    EdgeCollection,
    make_stub_patch as make_undirected_edge_patch,
)
from .edge.leaf import (
    LeafEdgeCollection,
)
from .label import (
    LabelCollection,
)
from .cascades import (
    CascadeCollection,
)
from .network import (
    _update_from_internal,
)


@_forwarder(
    (
        "set_clip_path",
        "set_clip_box",
        "set_snap",
        "set_sketch_params",
        "set_animated",
        "set_picker",
    )
)
class TreeArtist(mpl.artist.Artist):
    """Artist for plotting trees."""

    def __init__(
        self,
        tree,
        layout: Optional[str] = "horizontal",
        directed: bool = False,
        vertex_labels: Optional[bool | list[str] | dict[Hashable, str] | pd.Series] = None,
        edge_labels: Optional[Sequence | dict[Hashable, str] | pd.Series] = None,
        leaf_labels: Optional[Sequence | dict[Hashable, str]] | pd.Series = None,
        transform: mpl.transforms.Transform = mpl.transforms.IdentityTransform(),
        offset_transform: Optional[mpl.transforms.Transform] = None,
        show_support: bool = False,
    ):
        """Initialize the TreeArtist.

        Parameters:
            tree: The tree to plot.
            layout: The layout to use for the tree. Can be "horizontal", "vertical", or "radial".
            directed: Whether the tree is directed. Must be a boolean.
            vertex_labels: Labels for the vertices. Can be a list, dictionary, or pandas Series.
            edge_labels: Labels for the edges. Can be a sequence of strings.
            leaf_labels: Labels for the leaves. Can be a sequence of strings or a pandas Series.
                These labels are positioned at the depth of the deepest leaf. If you want to
                label leaves next to each leaf independently of how deep they are, use
                the "vertex_labels" parameter instead - usually as a dict with the leaves
                as keys and the labels as values.
            transform: The transform to apply to the tree artist. This is usually the identity.
            offset_transform: The offset transform to apply to the tree artist. This is
                usually `ax.transData`.
            show_support: Whether to show support values on the nodes. If both show_support and
                vertex_labels are set, this parameters takes precedence.
        """

        self.tree = tree
        self._ipx_internal_data = ingest_tree_data(
            tree,
            layout,
            directed=directed,
            layout_style=get_style(".layout", {}),
            vertex_labels=vertex_labels,
            edge_labels=edge_labels,
            leaf_labels=leaf_labels,
        )

        if show_support:
            support = self._ipx_internal_data["vertex_df"]["support"]
            self._ipx_internal_data["vertex_df"]["label"] = support

        super().__init__()

        # This is usually the identity (which scales poorly with dpi)
        self.set_transform(transform)

        # This is usually transData
        self.set_offset_transform(offset_transform)

        zorder = get_style(".network").get("zorder", 1)
        self.set_zorder(zorder)

        self._add_vertices()
        self._add_edges()
        self._add_leaf_vertices()
        self._add_leaf_edges()

        # NOTE: cascades need to be created after leaf vertices in case
        # they are requested to wrap around them.
        if get_style(".cascade") != {}:
            self._add_cascades()

    def get_children(self) -> tuple[mpl.artist.Artist]:
        """Get the children of this artist.

        Returns:
            The artists for vertices and edges.
        """
        children = [self._vertices, self._edges]
        if hasattr(self, "_leaf_vertices"):
            children.append(self._leaf_vertices)
        if hasattr(self, "_leaf_edges"):
            children.append(self._leaf_edges)
        if hasattr(self, "_cascades"):
            children.append(self._cascades)
        return tuple(children)

    def set_figure(self, fig) -> None:
        """Set the figure for this artist and its children.

        Parameters:
            fig: the figure to set for this artist and its children.
        """
        # At the end, if there are cadcades with extent depending on
        # leaf edges, we should update them
        super().set_figure(fig)

        # The next two are vanilla NetworkArtist
        self._vertices.set_figure(fig)
        self._edges.set_figure(fig)

        # For trees, there are a few more elements to coordinate,
        # including possibly text at the fringes (leaf labels)
        # which might require a redraw (without rendering) to compute
        # its actual scren real estate.
        if hasattr(self, "_leaf_vertices"):
            self._leaf_vertices.set_figure(fig)
        if hasattr(self, "_leaf_edges"):
            self._leaf_edges.set_figure(fig)
        if hasattr(self, "_cascades"):
            self._cascades.set_figure(fig)
        self._update_cascades_extent()

    def _update_cascades_extent(self) -> None:
        """Update cascades if extent depends on leaf labels."""
        if not hasattr(self, "_cascades"):
            return

        style_cascade = get_style(".cascade")
        extend_to_labels = style_cascade.get("extend", False) == "leaf_labels"
        if not extend_to_labels:
            return

        maxdepth = self._get_maxdepth_leaf_labels()
        self._cascades.set_maxdepth(maxdepth)

    def get_offset_transform(self):
        """Get the offset transform (for vertices/edges)."""
        return self._offset_transform

    def set_offset_transform(self, offset_transform):
        """Set the offset transform (for vertices/edges)."""
        self._offset_transform = offset_transform

    def get_layout(self, kind="vertex"):
        """Get vertex or edge layout."""
        layout_columns = [f"_ipx_layout_{i}" for i in range(self._ipx_internal_data["ndim"])]

        # Equivalence vertex <-> node
        if kind == "node":
            kind = "vertex"

        if kind == "vertex":
            layout = self._ipx_internal_data["vertex_df"][layout_columns]
            return layout
        elif kind == "leaf":
            leaves = self._ipx_internal_data["leaf_df"].index
            layout = self._ipx_internal_data["vertex_df"][layout_columns]
            # NOTE: workaround for a pandas bug
            idxs = []
            for i, vid in enumerate(layout.index):
                if vid in leaves:
                    idxs.append(i)
            layout = layout.iloc[idxs]
            return layout

        elif kind == "edge":
            return self._ipx_internal_data["edge_df"][layout_columns]
        else:
            raise ValueError(f"Unknown layout kind: {kind}. Use 'vertex' or 'edge'.")

    def get_datalim(self, transData, pad=0.15):
        """Get limits on x/y axes based on the graph layout data.

        Parameters:
            transData (Transform): The transform to use for the data.
            pad (float): Padding to add to the limits. Default is 0.05.
                Units are a fraction of total axis range before padding.
        """
        layout = self.get_layout().values

        if len(layout) == 0:
            return mpl.transforms.Bbox([[0, 0], [1, 1]])

        bbox = self._vertices.get_datalim(transData)

        edge_bbox = self._edges.get_datalim(transData)
        bbox = mpl.transforms.Bbox.union([bbox, edge_bbox])

        if hasattr(self, "_leaf_vertices"):
            leaf_labels_bbox = self._leaf_vertices.get_datalim(transData)
            bbox = mpl.transforms.Bbox.union([bbox, leaf_labels_bbox])

        if hasattr(self, "_cascades"):
            cascades_bbox = self._cascades.get_datalim(transData)
            bbox = mpl.transforms.Bbox.union([bbox, cascades_bbox])

        # NOTE: We do not need to check leaf edges for bbox, because they are
        # guaranteed within the convex hull of leaf vertices.

        bbox = bbox.expanded(sw=(1.0 + pad), sh=(1.0 + pad))
        return bbox

    def _get_label_series(self, kind: str) -> Optional[pd.Series]:
        # Equivalence vertex <-> node
        if kind == "node":
            kind = "vertex"

        if "label" in self._ipx_internal_data[f"{kind}_df"].columns:
            return self._ipx_internal_data[f"{kind}_df"]["label"]
        else:
            return None

    def get_vertices(self) -> VertexCollection:
        """Get VertexCollection artist."""
        return self._vertices

    get_nodes = get_vertices

    def get_edges(self) -> EdgeCollection:
        """Get EdgeCollection artist."""
        return self._edges

    def get_leaf_vertices(self) -> Optional[VertexCollection]:
        """Get leaf VertexCollection artist."""
        return self._leaf_vertices

    get_leaf_nodes = get_leaf_vertices

    def get_leaf_edges(self) -> Optional[LeafEdgeCollection]:
        """Get LeafEdgeCollection artist if present."""
        if hasattr(self, "_leaf_edges"):
            return self._leaf_edges
        return None

    def get_vertex_labels(self) -> LabelCollection:
        """Get list of vertex label artists."""
        return self._vertices.get_labels()

    get_node_labels = get_vertex_labels

    def get_edge_labels(self) -> LabelCollection:
        """Get list of edge label artists."""
        return self._edges.get_labels()

    def get_leaf_labels(self) -> Optional[LabelCollection]:
        """Get the leaf label artist if present."""
        return self._leaf_vertices.get_labels()

    def get_leaf_edge_labels(self) -> Optional[LabelCollection]:
        """Get the leaf edge label artist if present."""
        # TODO: leaf edge labels are basically unsupported as of now
        if hasattr(self, "_leaf_edges"):
            return self._leaf_edges.get_labels()
        return None

    def _add_vertices(self) -> None:
        """Add vertices to the tree."""
        self._vertices = VertexCollection(
            layout=self.get_layout(),
            layout_coordinate_system=self._ipx_internal_data.get(
                "layout_coordinate_system",
                "cartesian",
            ),
            style=get_style(".vertex"),
            labels=self._get_label_series("vertex"),
            transform=self.get_transform(),
            offset_transform=self.get_offset_transform(),
        )

    def _add_leaf_edges(self) -> None:
        """Add edges from the leaf to the max leaf depth."""
        # If there are no leaves, no leaf labels, or leaves are not deep,
        # skip leaf edges
        leaf_style = get_style(".leaf", {})
        if ("deep" not in leaf_style) and self.get_leaf_labels() is None:
            return
        if not leaf_style.get("deep", True):
            return

        # Given the conditions above, we should have leaf labels. If not,
        # make a None series with a valid index
        leaf_label_series = self._get_label_series("leaf")
        if leaf_label_series is None:
            leaf_label_series = self._ipx_internal_data["leaf_df"].copy()
            leaf_label_series["label"] = None
            leaf_label_series = leaf_label_series["label"]

        edge_style = get_style(
            ".leafedge",
        )
        default_style = {
            "linestyle": "--",
            "linewidth": 1,
            "color": "#111",
        }
        for key, value in default_style.items():
            if key not in edge_style:
                edge_style[key] = value

        labels = None
        # TODO: implement leaf edge labels
        # self._get_label_series("leafedge")

        if "cmap" in edge_style:
            cmap_fun = _build_cmap_fun(
                edge_style,
                "color",
            )
        else:
            cmap_fun = None

        if "cmap" in edge_style:
            colorarray = []
        edgepatches = []
        adjacent_vertex_ids = []
        for i, (vid, label) in enumerate(leaf_label_series.items()):
            # Use leaf label to compute backup key for style rotation
            # NOTE: This is quite a common use case. Users typically
            # refer to leaves via their labels because that's what you see
            # on screen. While an object exact match should take priority
            # for power users, a backup based on label is useful. Beginners
            # won't stumble upon this anyway since it's only used if
            # a dict-like property for leaf edges is provided, which is a
            # decently advanced thing to do.
            # NOTE: Multiple leaves might have the same label, in which case
            # all leaves will be style matched to this *unless* the dict-like
            # style object *also* matches on leaf objects directly, which
            # takes precedence since key2 is only a fallback.
            edge_stylei = rotate_style(edge_style, index=i, key=vid, key2=label)

            if cmap_fun is not None:
                colorarray.append(edge_stylei["color"])
                edge_stylei["color"] = cmap_fun(edge_stylei["color"])

            # These are not the actual edges drawn, only stubs to establish
            # the styles which are then fed into the dynamic, optimised
            # factory (the collection) below
            patch = make_undirected_edge_patch(
                **edge_stylei,
            )
            edgepatches.append(patch)
            adjacent_vertex_ids.append(vid)

        if "cmap" in edge_style:
            vmin = np.min(colorarray)
            vmax = np.max(colorarray)
            norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)
            edge_style["norm"] = norm

        self._leaf_edges = LeafEdgeCollection(
            edgepatches,
            vertex_leaf_ids=adjacent_vertex_ids,
            vertex_collection=self._vertices,
            leaf_collection=self._leaf_vertices,
            labels=labels,
            transform=self.get_offset_transform(),
            style=edge_style,
            directed=False,
        )
        if "cmap" in edge_style:
            self._leaf_edges.set_array(colorarray)

    def _add_leaf_vertices(self) -> None:
        """Add invisible deep vertices as leaf label anchors."""
        layout_name = self._ipx_internal_data["layout_name"]
        orientation = self._ipx_internal_data["orientation"]
        user_leaf_style = get_style(".leaf", {})

        leaf_layout = self.get_layout("leaf").copy()

        # Set all to max depth
        if user_leaf_style.get("deep", True):
            if layout_name == "radial":
                leaf_layout.iloc[:, 0] = leaf_layout.iloc[:, 0].max()
            elif (layout_name, orientation) == ("horizontal", "right"):
                leaf_layout.iloc[:, 0] = leaf_layout.iloc[:, 0].max()
            elif (layout_name, orientation) == ("horizontal", "left"):
                leaf_layout.iloc[:, 0] = leaf_layout.iloc[:, 0].min()
            elif (layout_name, orientation) == ("vertical", "descending"):
                leaf_layout.iloc[:, 1] = leaf_layout.iloc[:, 1].min()
            elif (layout_name, orientation) == ("vertical", "ascending"):
                leaf_layout.iloc[:, 1] = leaf_layout.iloc[:, 1].max()
            else:
                raise ValueError(
                    f"Layout and orientation not supported: {layout_name}, {orientation}."
                )

        # Set invisible vertices with visible labels
        if layout_name == "radial":
            ha = "auto"
            rotation = 0
        elif orientation == "right":
            ha = "left"
            rotation = 0
        elif orientation == "left":
            ha = "right"
            rotation = 0
        elif orientation == "ascending":
            ha = "left"
            rotation = 90
        else:
            ha = "left"
            rotation = -90

        default_leaf_style = {
            "size": 0,
            "label": {
                "verticalalignment": "center_baseline",
                "horizontalalignment": ha,
                "rotation": rotation,
                "hmargin": 5,
                "vmargin": 0,
                "bbox": {
                    "facecolor": (1, 1, 1, 0),
                },
            },
        }
        with context([{"vertex": default_leaf_style}, {"vertex": user_leaf_style}]):
            leaf_vertex_style = get_style(".vertex")
            # Left horizontal layout has no rotation of the labels but we need to
            # reverse hmargin
            if (
                layout_name == "horizontal"
                and orientation == "left"
                and "label" in leaf_vertex_style
                and "hmargin" in leaf_vertex_style["label"]
            ):
                # Reverse the horizontal margin
                leaf_vertex_style["label"]["hmargin"] *= -1

            self._leaf_vertices = VertexCollection(
                layout=leaf_layout,
                layout_coordinate_system=self._ipx_internal_data.get(
                    "layout_coordinate_system",
                    "catesian",
                ),
                style=leaf_vertex_style,
                labels=self._get_label_series("leaf"),
                transform=self.get_transform(),
                offset_transform=self.get_offset_transform(),
            )

    def _add_cascades(self) -> None:
        """Add cascade patches."""
        # NOTE: If leaf labels are present and the cascades are requested to wrap around them,
        # we have to compute the max extend of the cascades from the leaf labels.
        layout = self.get_layout()
        layout_name = self._ipx_internal_data["layout_name"]
        orientation = self._ipx_internal_data["orientation"]
        maxdepth = 1e-10
        if layout_name == "horizontal":
            if orientation == "right":
                maxdepth = layout.values[:, 0].max()
            else:
                maxdepth = layout.values[:, 0].min()
        elif layout_name == "vertical":
            if orientation == "descending":
                maxdepth = layout.values[:, 1].min()
            else:
                maxdepth = layout.values[:, 1].max()
        elif layout_name == "radial":
            # layout values are: r, theta
            maxdepth = layout.values[:, 0].max()

        style_cascade = get_style(".cascade")
        extend_to_labels = style_cascade.get("extend", False) == "leaf_labels"
        has_leaf_labels = self.get_leaf_labels() is not None
        if extend_to_labels and not has_leaf_labels:
            raise ValueError("Cannot extend cascades: no leaf labels.")

        if extend_to_labels and has_leaf_labels:
            maxdepth = self._get_maxdepth_leaf_labels()

        self._cascades = CascadeCollection(
            tree=self.tree,
            layout=layout,
            layout_name=layout_name,
            orientation=orientation,
            style=style_cascade,
            provider=data_providers["tree"][self._ipx_internal_data["tree_library"]],
            transform=self.get_offset_transform(),
            maxdepth=maxdepth,
        )

    def _get_maxdepth_leaf_labels(self):
        layout_name = self.get_layout_name()
        if layout_name == "radial":
            maxdepth = 0
            bboxes = self.get_leaf_labels().get_datalims_children(self.get_offset_transform())
            for bbox in bboxes:
                r1 = np.linalg.norm([bbox.xmax, bbox.ymax])
                r2 = np.linalg.norm([bbox.xmax, bbox.ymin])
                r3 = np.linalg.norm([bbox.xmin, bbox.ymax])
                r4 = np.linalg.norm([bbox.xmin, bbox.ymin])
                maxdepth = max(maxdepth, r1, r2, r3, r4)
        else:
            orientation = self.get_orientation()
            bbox = self.get_leaf_labels().get_datalim(self.get_offset_transform())
            if (layout_name, orientation) == ("horizontal", "right"):
                maxdepth = bbox.xmax
            elif layout_name == "horizontal":
                maxdepth = bbox.xmin
            elif (layout_name, orientation) == ("vertical", "descending"):
                maxdepth = bbox.ymin
            elif layout_name == "vertical":
                maxdepth = bbox.ymax

        return maxdepth

    def _add_edges(self) -> None:
        """Add edges to the network artist.

        NOTE: UndirectedEdgeCollection and ArrowCollection are both subclasses of
        PatchCollection. When used with a cmap/norm, they set their facecolor
        according to the cmap, even though most likely we only want the edgecolor
        set that way. It can make for funny looking plots that are not uninteresting
        but mostly niche at this stage. Therefore we sidestep the whole cmap thing
        here.
        """

        labels = self._get_label_series("edge")
        edge_style = get_style(".edge")

        edge_df = self._ipx_internal_data["edge_df"].set_index(["_ipx_source", "_ipx_target"])

        if "cmap" in edge_style:
            cmap_fun = _build_cmap_fun(
                edge_style,
                "color",
                internal=edge_df,
            )
        else:
            cmap_fun = None

        if "cmap" in edge_style:
            colorarray = []
        edgepatches = []
        adjacent_vertex_ids = []
        waypoints = []
        for i, (vid1, vid2) in enumerate(edge_df.index):
            edge_stylei = rotate_style(edge_style, index=i, key=vid2)

            # FIXME:: Improve this logic. We have three layers of priority:
            # 1. Explicitely set in the style of "plot"
            # 2. Internal through network attributes
            # 3. Default styles
            # Because 1 and 3 are merged as a style context on the way in,
            # it's hard to squeeze 2 in the middle. For now, we will assume
            # the priority order is 2-1-3 instead (internal property is
            # highest priority).
            # This is also why we cannot shift this logic further into the
            # EdgeCollection class, which is oblivious of NetworkArtist's
            # internal data. In fact, one would argue this needs to be
            # pushed outwards to deal with the wrong ordering.
            _update_from_internal(edge_stylei, edge_df.iloc[i], kind="edge")

            if cmap_fun is not None:
                colorarray.append(edge_stylei["color"])
                edge_stylei["color"] = cmap_fun(edge_stylei["color"])

            # Tree layout determines waypoints
            waypointsi = edge_stylei.pop("waypoints", None)
            if isinstance(waypointsi, (bool, np.bool)):
                waypointsi = ["none", None][int(waypointsi)]
            if waypointsi is None:
                layout_name = self._ipx_internal_data["layout_name"]
                if layout_name == "horizontal":
                    waypointsi = "x0y1"
                elif layout_name == "vertical":
                    waypointsi = "y0x1"
                elif layout_name == "radial":
                    waypointsi = "r0a1"
                # NOTE: no need to catch the default case, it's caught
                # when making the layout already. We should *never* be
                # in an "else" case here.
            waypoints.append(waypointsi)

            # These are not the actual edges drawn, only stubs to establish
            # the styles which are then fed into the dynamic, optimised
            # factory (the collection) below
            patch = make_undirected_edge_patch(
                **edge_stylei,
            )
            edgepatches.append(patch)
            adjacent_vertex_ids.append((vid1, vid2))

        if "cmap" in edge_style:
            vmin = np.min(colorarray)
            vmax = np.max(colorarray)
            norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)
            edge_style["norm"] = norm

        if get_style(".layout", {}).get("angular", False):
            edge_style.pop("waypoints", None)
        else:
            edge_style["waypoints"] = waypoints

        self._edges = EdgeCollection(
            edgepatches,
            vertex_ids=adjacent_vertex_ids,
            vertex_collection=self._vertices,
            labels=labels,
            transform=self.get_offset_transform(),
            style=edge_style,
            directed=bool(self._ipx_internal_data["directed"]),
        )
        if "cmap" in edge_style:
            self._edges.set_array(colorarray)

    def get_layout_name(self) -> str:
        """Get the layout name."""
        return self._ipx_internal_data["layout_name"]

    def get_orientation(self) -> Optional[str]:
        """Get the orientation of the tree layout."""
        return self._ipx_internal_data.get("orientation", None)

    def style_subtree(
        self,
        nodes: Sequence[Hashable],
        style: Optional[dict[str, Any] | Sequence[str | dict[str, Any]]] = None,
        **kwargs,
    ) -> None:
        """Style a subtree of the tree.

        Parameters:
            nodes: Sequence of nodes that span the subtree. All elements below including
                the most recent common ancestor of these leaves will be styled.
            style: Style or sequence of styles to apply to the subtree. Each style can
                be either a string, referring to an internal `iplotx` style, or a dictionary
                with custom styling elements.
            kwargs: Additional flat style elements. If both style and kwargs are provided,
                kwargs is applied last.
        """
        styles = []
        if isinstance(style, (str, dict)):
            styles = [style]
        elif style is not None:
            styles = list(style)
        style = merge_styles(styles + [kwargs])

        provider = data_providers["tree"][self._ipx_internal_data["tree_library"]]

        # Get last (deepest) common ancestor of the requested nodes
        root = provider(self.tree).get_lca(nodes)

        # Populate a DataFrame with the array of properties to update
        vertex_idx = {node: i for i, node in enumerate(self._ipx_internal_data["vertex_df"].index)}
        edge_idx = {
            node: i
            for i, node in enumerate(self._ipx_internal_data["edge_df"]["_ipx_target"].values)
        }
        vertex_props = {}
        edge_props = {}
        vertex_style = style.get("vertex", {})
        edge_style = style.get("edge", {})
        for inode, node in enumerate(provider(root).preorder()):
            for attr, value in vertex_style.items():
                if attr not in vertex_props:
                    vertex_props[attr] = list(getattr(self._vertices, f"get_{attr}")())
                vertex_props[attr][vertex_idx[node]] = value

            # Ignore branch coming into the root node
            if inode == 0:
                continue

            for attr, value in edge_style.items():
                # Edge color is actually edgecolor
                if attr == "color":
                    attr = "edgecolor"
                if attr not in edge_props:
                    edge_props[attr] = list(getattr(self._edges, f"get_{attr}")())
                edge_props[attr][edge_idx[node]] = value

        # Update the properties from the DataFrames
        for attr in vertex_props:
            getattr(self._vertices, f"set_{attr}")(vertex_props[attr])
        for attr in edge_props:
            getattr(self._edges, f"set_{attr}")(edge_props[attr])

    @_stale_wrapper
    def draw(self, renderer) -> None:
        """Draw each of the children, with some buffering mechanism."""
        if not self.get_visible():
            return

        # NOTE: looks like we have to manage the zorder ourselves
        # this is kind of funny actually. Btw we need to ensure
        # that cascades are drawn behind (earlier than) vertices
        # and edges at equal zorder because it looks better that way.
        z_suborder = defaultdict(int)
        if hasattr(self, "_leaf_vertices"):
            z_suborder[self._leaf_vertices] = -2
        if hasattr(self, "_leaf_edges"):
            z_suborder[self._leaf_edges] = -2
        if hasattr(self, "_cascades"):
            z_suborder[self._cascades] = -1
        children = list(self.get_children())
        children.sort(key=lambda x: (x.zorder, z_suborder[x]))
        for art in children:
            if isinstance(art, CascadeCollection):
                self._update_cascades_extent()
            art.draw(renderer)



"""
Subtree styling
===============

This example shows how to style tree clades - or subtrees - post rendering,
similar of the equivalent option in `cogent3`.
"""

import cogent3
import matplotlib.pyplot as plt
import iplotx as ipx

tree = cogent3.load_tree("data/tree-with-support.json")

fig, ax = plt.subplots(figsize=(8, 8))
art = ipx.tree(
    tree,
    layout="radial",
    ax=ax,
    leaf_labels=True,
    layout_angular=True,
    leaf_deep=False,
    margin=0.1,
)

# Style the subtree spanned by the first two leaves
# with red edges.
art.style_subtree(
    tree.tips()[:2],
    {
        "edge": {
            "color": "red",
        },
    },
)


# %%
# The same can be done with nonangular layouts:

fig, ax = plt.subplots(figsize=(8, 8))
art = ipx.tree(
    tree,
    layout="horizontal",
    ax=ax,
    leaf_labels=True,
    layout_angular=False,
    leaf_deep=False,
    margin=0.1,
)
art.style_subtree(
    tree.tips()[:2],
    {
        "edge": {
            "color": "red",
        },
    },
)



from typing import (
    Optional,
    Sequence,
)
from collections.abc import Hashable
import importlib
import numpy as np
import pandas as pd

from ....typing import (
    LayoutType,
)
from ...heuristics import (
    normalise_layout,
)
from ...typing import (
    NetworkDataProvider,
    NetworkData,
)
from ....utils.internal import (
    _make_layout_columns,
)


class NetworkXDataProvider(NetworkDataProvider):
    def __call__(
        self,
        layout: Optional[LayoutType] = None,
        vertex_labels: Optional[Sequence[str] | dict[Hashable, str] | pd.Series] = None,
        edge_labels: Optional[Sequence[str] | dict[str]] = None,
    ) -> NetworkData:
        """Create network data object for iplotx from a networkx object."""

        import networkx as nx

        # Get layout
        vertex_df = normalise_layout(
            layout,
            network=self.network,
            nvertices=self.number_of_vertices(),
        )
        ndim = vertex_df.shape[1]
        vertex_df.columns = _make_layout_columns(ndim)

        # Vertices are indexed by node ID
        vertex_df = vertex_df.loc[pd.Index(self.network.nodes)]

        # Vertex internal properties
        tmp = pd.DataFrame(dict(self.network.nodes.data())).T
        # Arrays become a single column, which we have already anyway
        if isinstance(layout, str) and (layout in tmp.columns):
            del tmp[layout]
        for col in tmp.columns:
            vertex_df[col] = tmp[col]
        del tmp

        # Vertex labels
        # Recast vertex_labels=False as vertex_labels=None
        if np.isscalar(vertex_labels) and (not vertex_labels):
            vertex_labels = None
        if vertex_labels is None:
            if "label" in vertex_df:
                del vertex_df["label"]
        else:
            if np.isscalar(vertex_labels) and (not vertex_labels) and ("label" in vertex_df):
                del vertex_df["label"]
            elif vertex_labels is True:
                if "label" not in vertex_df:
                    vertex_df["label"] = vertex_df.index
            elif (not np.isscalar(vertex_labels)) and (len(vertex_labels) != len(vertex_df)):
                raise ValueError("Vertex labels must be the same length as the number of vertices.")
            elif isinstance(vertex_labels, nx.classes.reportviews.NodeDataView):
                vertex_df["label"] = pd.Series(dict(vertex_labels))
            else:
                vertex_df["label"] = vertex_labels

        # Edges are a list of tuples, because of multiedges
        tmp = []
        for u, v, d in self.network.edges.data():
            row = {"_ipx_source": u, "_ipx_target": v}
            row.update(d)
            tmp.append(row)
        if len(tmp):
            edge_df = pd.DataFrame(tmp)
        else:
            edge_df = pd.DataFrame(columns=["_ipx_source", "_ipx_target"])
        del tmp

        # Edge labels
        # Even though they could exist in the dataframe, request the to be explicitely mentioned
        if (edge_labels is None) and ("label" in edge_df):
            del edge_df["label"]
        elif edge_labels is not None:
            if np.isscalar(edge_labels):
                if (not edge_labels) and ("label" in edge_df):
                    del edge_df["label"]
                if (edge_labels is True) and ("label" not in edge_df):
                    edge_df["label"] = [str(i) for i in edge_df.index]
            else:
                if len(edge_labels) != len(edge_df):
                    raise ValueError("Edge labels must be the same length as the number of edges.")
                if isinstance(edge_labels, dict):
                    edge_labels = pd.Series(edge_labels)
                    edge_df["label"] = edge_labels.loc[
                        edge_df.set_index(["_ipx_source", "_ipx_target"]).index
                    ].values
                else:
                    edge_df["label"] = edge_labels

        network_data = {
            "vertex_df": vertex_df,
            "edge_df": edge_df,
            "directed": self.is_directed(),
            "ndim": ndim,
        }
        return network_data

    @staticmethod
    def check_dependencies() -> bool:
        return importlib.util.find_spec("networkx") is not None

    @staticmethod
    def graph_type():
        from networkx import Graph

        return Graph

    def is_directed(self):
        import networkx as nx

        return isinstance(self.network, (nx.DiGraph, nx.MultiDiGraph))

    def number_of_vertices(self):
        """The number of vertices/nodes in the network."""
        return self.network.number_of_nodes()



import pytest
import iplotx as ipx


def test_incomplete_network_protocol():
    class IncompleteProtocol(ipx.ingest.typing.NetworkDataProvider):
        pass

    protocol = IncompleteProtocol(None)

    with pytest.raises(NotImplementedError):
        protocol()

    with pytest.raises(NotImplementedError):
        protocol.check_dependencies()

    with pytest.raises(NotImplementedError):
        protocol.graph_type()

    with pytest.raises(NotImplementedError):
        protocol.is_directed()

    with pytest.raises(NotImplementedError):
        protocol.number_of_vertices()


def test_incomplete_tree_protocol():
    class IncompleteProtocol(ipx.ingest.typing.TreeDataProvider):
        pass

    protocol = IncompleteProtocol(None)

    with pytest.raises(NotImplementedError):
        protocol.check_dependencies()

    with pytest.raises(NotImplementedError):
        protocol.tree_type()

    with pytest.raises(NotImplementedError):
        protocol.get_leaves()

    with pytest.raises(NotImplementedError):
        protocol.preorder()

    with pytest.raises(NotImplementedError):
        protocol.postorder()

    with pytest.raises(NotImplementedError):
        protocol.get_children(None)

    with pytest.raises(NotImplementedError):
        protocol.get_branch_length(None)



style_library = {
    "default": {
        "vertex": {
            "size": 20,
            "facecolor": "black",
            "marker": "o",
            "label": {
                "color": "white",
                "horizontalalignment": "center",
                "verticalalignment": "center",
                "hpadding": 18,
                "vpadding": 12,
            },
        },
        "edge": {
            "linewidth": 1.5,
            "linestyle": "-",
            "color": "black",
            "curved": False,
            "tension": 1,
            "looptension": 4,
            "loopmaxangle": 60,
            "paralleloffset": 3,
            "label": {
                "horizontalalignment": "center",
                "verticalalignment": "center",
                "rotate": False,
                "bbox": {
                    "boxstyle": "round",
                    "facecolor": "white",
                    "edgecolor": "none",
                },
            },
            "arrow": {
                "marker": "|>",
                "width": 8,
            },
        },
        "grouping": {
            "facecolor": ["grey", "steelblue", "tomato"],
            "edgecolor": "black",
            "linewidth": 1.5,
            "alpha": 0.5,
            "vertexpadding": 18,
        },
    },
    # Hollow style for organization charts et similar
    "hollow": {
        "vertex": {
            "color": None,
            "facecolor": "none",
            "edgecolor": "black",
            "linewidth": 1.5,
            "marker": "r",
            "size": "label",
            "label": {
                "color": "black",
            },
        }
    },
    # Feedback, for regulatory networks
    "feedback": {
        "edge": {
            "linewidth": 4,
            "shrink": 10,
            "arrow": {
                "marker": ")>",
                "width": 20,
                "height": 28,
            },
        },
        "vertex": {
            "size": 35,
            "color": None,
            "facecolor": "white",
            "edgecolor": "black",
            "linewidth": 4,
            "label": {
                "color": "black",
            },
        },
    },
    # Tree style, with zero-size vertices
    "tree": {
        "vertex": {
            "size": 0,
            "label": {
                "color": "black",
                "size": 12,
                "verticalalignment": "center",
                "hmargin": 10,
                "bbox": {
                    "boxstyle": "square,pad=0.5",
                    "facecolor": "none",
                    "edgecolor": "none",
                },
            },
        },
        "edge": {
            "linewidth": 2.5,
        },
    },
    # Cogent3 tree style
    "cogent3": {
        "vertex": {
            "size": 3,
            "label": {
                "color": "black",
                "size": 10,
                "verticalalignment": "center",
                "bbox": {
                    "facecolor": "none",
                    "edgecolor": "none",
                },
            },
        },
        "edge": {
            "linewidth": 1.5,
        },
        "leaf": {
            "deep": False,
        },
    },
    # Dashed depth tree branches
    "dashdepth": {
        "vertex": {
            "size": 2,
            "label": {
                "color": "black",
                "size": 10,
                "verticalalignment": "center",
                "bbox": {
                    "facecolor": "none",
                    "edgecolor": "none",
                },
            },
        },
        "edge": {
            "linewidth": 1.5,
            "split": {
                "linestyle": ":",
            },
        },
        "leaf": {
            "deep": False,
        },
    },
    # Dashed depth tree branches
    "dashwidth": {
        "vertex": {
            "size": 2,
            "label": {
                "color": "black",
                "size": 10,
                "verticalalignment": "center",
                "bbox": {
                    "facecolor": "none",
                    "edgecolor": "none",
                },
            },
        },
        "edge": {
            "linewidth": 1.5,
            "linestyle": ":",
            "split": {
                "linestyle": "-",
            },
        },
        "leaf": {
            "deep": False,
        },
    },
    # Greyscale style
    "greyscale": {
        "vertex": {
            "color": None,
            "facecolor": "lightgrey",
            "edgecolor": "#111",
            "marker": ">",
            "size": 15,
            "linewidth": 0.75,
            "label": {
                "color": "black",
            },
        },
        "edge": {
            "color": "#111",
            "linewidth": 0.75,
            "arrow": {
                "marker": ">",
            },
            "label": {
                "rotate": True,
                "color": "black",
            },
        },
    },
    # Edge highlight
    "rededge": {
        "vertex": {
            "size": 13,
            "color": None,
            "facecolor": "none",
            "edgecolor": "#111",
            "linewidth": 2,
            "label": {
                "color": "#111",
            },
        },
        "edge": {
            "color": "tomato",
            "linewidth": 2.5,
        },
    },
    # Vertex highlight
    "rednode": {
        "vertex": {
            "size": 22,
            "color": None,
            "facecolor": "tomato",
            "edgecolor": "firebrick",
            "linewidth": 2,
        },
        "edge": {
            "color": "#333",
            "linewidth": 1,
        },
    },
    # Eerie style
    "eerie": {
        "vertex": {
            "color": None,
            "facecolor": "white",
            "edgecolor": "#111",
            "linestyle": "--",
            "linewidth": 2,
        },
        "edge": {
            "color": "#111",
            "linestyle": "--",
            "linewidth": 2,
            "arrow": {
                "marker": ")",
            },
        },
    },
    # Emulate a little networkx
    "networkx": {
        "vertex": {
            "color": None,
            "facecolor": "steelblue",
            "edgecolor": "none",
            "label": {
                "color": "black",
            },
        },
        "edge": {
            "color": "black",
            "linewidth": 1.5,
            "arrow": {
                "width": 5,
            },
        },
    },
    "igraph": {
        "vertex": {
            "color": None,
            "facecolor": "lightblue",
            "edgecolor": "black",
            "linewidth": 2,
            "label": {
                "color": "black",
            },
        },
        "edge": {
            "color": "black",
            "linewidth": 2,
            "arrow": {
                "marker": "|>",
                "width": 6.5,
            },
        },
    },
    # Colorful style for general use
    "unicorn": {
        "vertex": {
            "size": 23,
            "color": None,
            "facecolor": [
                "darkorchid",
                "slateblue",
                "seagreen",
                "lime",
                "gold",
                "orange",
                "sandybrown",
                "tomato",
                "deeppink",
            ],
            "edgecolor": "black",
            "linewidth": 1,
            "marker": "*",
            "label": {
                "color": [
                    "white",
                    "white",
                    "white",
                    "black",
                    "black",
                    "black",
                    "black",
                    "white",
                    "white",
                ],
            },
        },
        "edge": {
            "color": [
                "darkorchid",
                "slateblue",
                "seagreen",
                "lime",
                "gold",
                "orange",
                "sandybrown",
                "tomato",
                "deeppink",
            ],
            "linewidth": 1.5,
            "label": {
                "rotate": False,
                "color": [
                    "white",
                    "white",
                    "white",
                    "black",
                    "black",
                    "black",
                    "black",
                    "white",
                    "white",
                ],
            },
        },
    },
}



"""
Charts as graph visualisations
==============================

This example shows how to use ``iplotx`` to display changes of a quantity over time.
This is a kind of chart that would not usually be considered a graph, but can be
reinterpreted that way to make use of ``iplotx``'s visual capabilities.
"""

import datetime
import networkx as nx
import matplotlib.pyplot as plt
import iplotx as ipx

times = ["2021-01-01", "2021-02-01", "2021-03-01", "2021-04-01", "2021-05-01"]
quantity = [10, 15, 7, 20, 12]

# Convert the date strings to numbers because iplotx does not understand dates
dates = [datetime.datetime(*list(map(int, t.split("-")))) for t in times]
days_since_start = [(d - dates[0]).days for d in dates]

g = nx.path_graph(len(quantity), create_using=nx.DiGraph)
layout = {i: (days_since_start[i], quantity[i]) for i in range(len(quantity))}

fig, ax = plt.subplots()
ipx.network(
    g,
    layout,
    ax=ax,
    strip_axes=False,
    zorder=2,
)
ax.set_xticks(days_since_start)
ax.set_xticklabels(times, rotation=45, ha="right")
ax.set_ylabel("Quantity")
ax.grid(True)
fig.tight_layout()

# %%
# This can also be done in another way by using a single edge:

g = nx.DiGraph([(0, 4)])
fig, ax = plt.subplots()
ipx.network(
    g,
    layout,
    ax=ax,
    strip_axes=False,
    zorder=2,
    edge_waypoints=[[
        layout[1],
        layout[2],
        layout[3],
    ]],
)
ax.set_xticks(days_since_start)
ax.set_xticklabels(times, rotation=45, ha="right")
ax.set_ylabel("Quantity")
ax.grid(True)
fig.tight_layout()



"""
Social circles
==============

This example inspired by SocNetV (https://socnetv.org/data/uploads/screenshots/32/socnetv-32-example-dot.png), shows social connections within circles.
It also demonstrates how to combine ``iplotx`` with other matplotlib artists.

"""

import igraph as ig
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import iplotx as ipx


edge_data = [
    (1, 2),
    (2, 3),
    (3, 4),
    (4, 5),
    (5, 1),
    (1, 6),
    (6, 7),
    (7, 8),
    (8, 9),
    (9, 10),
]
edge_data = pd.DataFrame(edge_data).astype(str)

g = ig.Graph.DataFrame(
    edge_data,
    use_vids=False,
    directed=True,
)


def coords(radius, angle):
    angle = np.deg2rad(angle)
    return radius * np.cos(angle), radius * np.sin(angle)


radii = [1, 3, 5, 6, 7, 12]
layout = [
    coords(radii[0], 0),
    coords(radii[4], -30),
    coords(radii[3], -70),
    coords(radii[2], -100),
    coords(radii[1], -135),
    coords(radii[5], 180),
    coords(radii[5], 180 - 36),
    coords(radii[5], 180 - 36 * 2),
    coords(radii[5], 180 - 36 * 3),
    coords(radii[5], 180 - 36 * 4),
]


fig, ax = plt.subplots(figsize=(8, 8))
ipx.network(
    g,
    layout,
    ax=ax,
    vertex_marker="s",
    vertex_facecolor="red",
    edge_shrink=3,
    edge_zorder=2,
    vertex_zorder=3,
    vertex_label_color="black",
    vertex_labels=list(g.vs["name"]),
    zorder=3,
)

for radius in radii:
    ax.add_patch(
        plt.Circle(
            (0, 0), radius,
            facecolor="none", edgecolor="red", ls="dashed",
            zorder=1,
        )
    )

ax.set(xlim=(-13, 13), ylim=(-13, 13))
fig.tight_layout()



"""
Subtrees and cascades
=====================

This Biopython-inspired example shows how to style clade edges for a
node and its descendants.

"""

from Bio import Phylo
from io import StringIO
import iplotx as ipx

tree = Phylo.read(
    StringIO("(((A,B),(C,D)),(E,F,G));"),
    format="newick",
)

mrca = tree.common_ancestor({"name": "E"}, {"name": "F"})

art = ipx.tree(
    tree,
    leaf_deep=True,
    leaf_labels=True,
    cascade_facecolor={
        mrca: "salmon",
        tree.clade[0, 1]: "steelblue",
    },
    cascade_extend=True,
)
art.style_subtree(
    [mrca],
    {"edge": {"color": "purple"}},
)
art.style_subtree(
    [tree.clade[0, 1]],
    {"edge": {"color": "navy"}},
)



"""
Animal phylogeny
================

This example from ``cogent3`` shows how to use ``iplotx`` to visualise a phylogenetic tree of many animals.

It also shows how to combine ``iplotx`` trees with other ``matplotlib`` artists such as annotations and
scatter plots.
"""

from collections import defaultdict
import cogent3
import matplotlib.pyplot as plt
import iplotx as ipx

reader = cogent3.get_app("load_json")

ens_tree = reader("data/GN-tree.json")

# Customise the figure as you like
fig, ax = plt.subplots(figsize=(8, 14))

# Inject plot into the figure/axes
tree_artist = ipx.tree(
  ens_tree,
  layout="horizontal",
  ax=ax,
  leaf_labels=True,
  # Style options
  layout_angular=False,
  leaf_deep=True,
  margins=(0.2, 0),
  leafedge_color=defaultdict(lambda: "black", {
      "Human": "tomato",
      "Chimpanzee": "orange",
      "Orangutan": "gold",
      "Gorilla": "gold",
      "Rhesus": "yellow",
      "HowlerMon": "yellow",
  }),
  leafedge_linewidth=2,
)

# Add an annotation with an arrow towards the root
layout = tree_artist.get_layout().values
root_coords = layout[layout[:, 0] == 0][0]
ax.annotate(
    "Tree root",
    root_coords,
    (-0.1, 55),
    xycoords="data",
    textcoords="data",
    arrowprops=dict(
      color="grey",
      arrowstyle="-|>",
      shrinkA=4,
      shrinkB=12,
      linewidth=2,
      connectionstyle="angle",
    ),
    bbox=dict(
        boxstyle="round,rounding_size=0.2,pad=0.5",
        facecolor="white",
        edgecolor="grey",
        linewidth=2,
    ),
    fontsize=12,
)

# Also add concentric circles at the root
ax.scatter(
    [root_coords[0]] * 3,
    [root_coords[1]] * 3,
    s=[50, 200, 500],
    facecolor="none",
    edgecolor="orchid",
    linewidth=2,
)

# Ensure tight layout for minimal whitespace
fig.tight_layout()



"""
TCA cycle
=========

This example visualises a subset of the tricarboxylic acid (TCA) cycle, also known as the Krebs cycle or citric acid cycle.

"""

from collections import defaultdict
import numpy as np
import pandas as pd
import networkx as nx
import matplotlib.pyplot as plt
import iplotx as ipx



g = nx.DiGraph()
g.add_edges_from([
    ("oxaloacetate", "citrate"),
    ("citrate", "isocitrate"),
    ("isocitrate", "alpha-ketoglutarate"),
    ("alpha-ketoglutarate", "succinyl-CoA"),
    ("succinyl-CoA", "succinate"),
    ("succinate", "fumarate"),
    ("fumarate", "malate"),
    ("malate", "oxaloacetate"),
])
edge_labels = {
    (u, v): str(i+1) for i, (u, v) in enumerate(g.edges)
}


layout = pd.DataFrame(nx.circular_layout(g)).T
layout.values[:] = layout.values @ np.array([[0, -1], [1, 0]])
layout[1] *= -1

g.add_edges_from([
    ("acetyl CoA", "CoA-SH"),
    ("NAD+", "NADH"),
    ("NAD+ ", "NADH "),
    ("NAD+  ", "NADH  "),
    ("GDP", "GTP"),
    ("GTP", "GDP"),
    ("ADP", "ATP"),
])

layout.loc["acetyl CoA"] = [0.27, 1.18]
layout.loc["CoA-SH"] = [0.75, 1.05]
layout.loc["NAD+"] = [0.68, -0.15]
layout.loc["NADH"] = [0.61, -0.45]
layout.loc["NAD+ "] = [0.65, -1.03]
layout.loc["NADH "] = [0.35, -1.2]
layout.loc["NAD+  "] = [-0.65, 1.03]
layout.loc["NADH  "] = [-0.35, 1.2]
layout.loc["GDP"] = [-0.25, -1.18]
layout.loc["GTP"] = [-0.7, -0.98]
layout.loc["ATP"] = [-0.35, -1.45]
layout.loc["ADP"] = [-0.8, -1.25]

edge_tension = {
    ("oxaloacetate", "citrate"): -0.8,
    ("citrate", "isocitrate"): -0.8,
    ("isocitrate", "alpha-ketoglutarate"): -0.8,
    ("alpha-ketoglutarate", "succinyl-CoA"): -0.8,
    ("succinyl-CoA", "succinate"): -0.8,
    ("succinate", "fumarate"): -0.8,
    ("fumarate", "malate"): -0.8,
    ("malate", "oxaloacetate"): -0.8,
    ("acetyl CoA", "CoA-SH"): 5.7,
    ("NAD+", "NADH"): -10.0,
    ("NAD+ ", "NADH "): 8.0,
    ("NAD+  ", "NADH  "): 8.0,
    ("GDP", "GTP"): 4.0,
    ("GTP", "GDP"): 3.2,
    ("ADP", "ATP"): -3.2,
}

edge_color = defaultdict(
    lambda: "black",
    {
        ("acetyl CoA", "CoA-SH"): "tomato",
        ("ADP", "ATP"): "gold",
    },
)

vertex_facecolor = defaultdict(
    lambda: "white",
    {
        "acetyl CoA": "tomato",
        "CoA-SH": "tomato",
        "ADP": "gold",
        "ATP": "gold",
    },
)

for (u, v) in g.edges:
    if (u, v) not in edge_labels:
        edge_labels[(u, v)] = ""


# Visualise
fig, ax = plt.subplots(figsize=(7, 8.5))
ipx.network(
    g,
    layout=layout,
    ax=ax,
    style="hollow",
    vertex_labels=True,
    edge_curved=True,
    edge_tension=edge_tension,
    edge_labels=edge_labels,
    edge_label_rotate=True,
    edge_color=edge_color,
    vertex_facecolor=vertex_facecolor,
    vertex_alpha=0.7,
)
fig.tight_layout()



"""
Arrowheads
==========

This example showcases various types of arrowheads for connected graphs.
"""

import networkx as nx
import matplotlib.pyplot as plt
import iplotx as ipx

arrow_markers = ["|>", "|/", "|\\", ">", "<", ">>", ")>", ")", "(", "]", "[", "|",
                 "x", "s", "d", "p", "q"]
n = len(arrow_markers)
G = nx.DiGraph()
G.add_edges_from([(f"l{i}", f"r{i}") for i in range(n)])
layout = {f"l{i}": (0, -i) for i in range(n)}
layout.update({f"r{i}": (1, -i) for i in range(n)})

fig, ax = plt.subplots(figsize=(3, 8.5))
ipx.network(
    G,
    layout=layout,
    ax=ax,
    vertex_size=12,
    edge_arrow_marker=arrow_markers,
    edge_arrow_height="width",
    edge_shrink=8,
    edge_color="steelblue",
    title="Arrow markers",
)
ax.set_yticks(list(range(-n + 1, 1))[::-1])
ax.set_yticklabels(arrow_markers)
ax.yaxis.set_tick_params(length=0)



"""
Complex example
===============

This example from networkx shows how to use complex network properties including edge tension.
"""

import networkx as nx
import matplotlib as mpl
import matplotlib.pyplot as plt
import iplotx as ipx
import itertools as it

nodes = "ABC"
prod = list(it.product(nodes, repeat=2)) * 4
G = nx.MultiDiGraph()
for i, (u, v) in enumerate(prod):
    G.add_edge(u, v, w=round(i / 3, 2))
nx.set_node_attributes(G, nx.spring_layout(G, seed=3113794652), "pos")
csi = it.cycle([5 * r for r in it.accumulate([0.15] * 4)])
nx.set_edge_attributes(G, {e: next(csi) for e in G.edges(keys=True)}, "tension")
nx.set_edge_attributes(
    G,
    {tuple(e): w for *e, w in G.edges(keys=True, data="w")},
    "label",
)

fig, ax = plt.subplots()
ipx.plot(
    G,
    ax=ax,
    layout="pos",
    edge_labels=True,
    style={
        "edge": {
            "curved": True,
            "tension": G.edges.data("tension"),
            "color": G.edges.data("w"),
            "cmap": mpl.colormaps["inferno"],
            "linewidth": 1,
            "looptension": 7.5,
            "label": {
                "color": "black",
                "bbox": {
                    "facecolor": "none",
                },
            },
        },
    },
)
ax.margins(0.17)



"""
Financial network
=================

This example demonstrates the use of ``iplotx`` to visualise financial relationships, mimicking a figure published in this paper: https://www.science.org/doi/10.1126/science.1173644.
"""

from collections import defaultdict
import igraph as ig
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import iplotx as ipx


# Circle nodes
g = ig.Graph(23, directed=True)
layout = g.layout_circle().coords
names = [
    "Commerzbank",
    "ING",
    "Prudential fin.",
    "HBOS",
    "Lloyds TSB",
    "Mitsubishi UFJ",
    "Generali",
    "Sumitomo",
    "Royal Bank Scotland",
    "Gen. Electric",
    "Intesa-Sanpaolo",
    "Aberdeen",
    "Mediobanca",
    "Sumitomomitsui",
    "Fidelity Mng.",
    "IFI",
    "Deutsche Bank",
    "Franklin Res.",
    "Merril Lynch",
    "Wellington Mng.",
    "Nomura",
    "Barclays",
    "Unicredito",
]

# Internal nodes
data_internal = [
    ["Bank Nova Scotia", -0.1, 0.8],
    ["Bear Stearns", -0.4, 0.6],
    ["UBS", -0.7, 0.3],
    ["Friends Provident", -0.4, -0.4],
    ["Cr. Suisse", -0.1, -0.6],
    ["Soc. Generale", -0.3, 0.2],
    ["FMR Corp", -0.15, -0.05],
    ["Bank of America", 0.25, 0.7],
    ["Santander", 0.1, 0.4],
    ["Citigroup", 0.6, -0.3],
    ["BNP Paribas", 0.6, -0.5],
    ["HSBC", 0.35, -0.2],
    ["JP Morgan Chase", 0.22, -0.6],
    ["Morgan Stanley", 0.6, 0.2],
    ["Goldman Sachs", 0.5, 0.5],
    ["Capital Group", 0.3, 0.1],
    ["Key Corp", 0.1, -0.35],
]
layout.extend([x[1:] for x in data_internal])
names.extend([x[0] for x in data_internal])
g.add_vertices(len(layout) - g.vcount())


g.vs["name"] = names

g.add_edges([
    (0, 1),
    (1, 2),
    (2, 3),
    (22, 0),
    (26, 2),
    (20, 0),
    (20, 1),
    (20, 18),
    (22, 2),
    (22, 3),
    (22, 4),
    (8, 5),
    (8, 9),
    (8, 10),
    (6, 11),
    (11, 6),
    (8, 12),
    (8, 13),
    (8, 23),
    (8, 24),
    (8, 27),
    (8, 31),
    (8, 37),
    (12, 13),
    (12, 14),
    (15, 14),
    (20, 15),
    (16, 17),
    (18, 16),
    (17, 16),
    (18, 19),
    (27, 35),
    (20, 35),
    (18, 35),
    (34, 33),
    (22, 36),
    (38, 21),
    (12, 25),
    (22, 25),
    (28, 38),
    (26, 13),
    (30, 5),
    (4, 30),
    (30, 4),
    (22, 30),
    (37, 4),
    (37, 12),
    (37, 14),
    (37, 15),
    (26, 15),
    (26, 14),
    (26, 24),
    (32, 29),
    (17, 27),
    (27, 17),
    (18, 27),
    (27, 18),
    (19, 18),
    (11, 39),
    (6, 39),
    (8, 35),
    (8, 19),
    (2, 34),
    (2, 32),
    (2, 33),
    (2, 20),
    (1, 34),
    (30, 12),
    (19, 35),
    (26, 20),
    (26, 33),
    # weight 2
    (8, 26),
    # weight 3
    (16, 27),
    (16, 35),
    (16, 18),
    (22, 32),
    (22, 33),
    # weight 4
    (8, 7),
    (27, 16),
    (24, 25),
    (22, 5),
    (22, 37),
    # weight 5
    (23, 6),
    (37, 30),
])

g.es["weight"] = 1
g.es[-2 -5 -5-1:]["weight"] = 3
g.es[-2 -5 -5:]["weight"] = 3
g.es[-2 -5:]["weight"] = 4
g.es[-2:]["weight"] = 5
edge_colord = {
    1: "goldenrod",
    2: "darkgoldenrod",
    3: "peru",
    4: "saddlebrown",
    5: "black",
}
g.es["color"] = [edge_colord[w] for w in g.es["weight"]]
g.es["linewidth"] = [0.5 + 0.8 * w for w in g.es["weight"]]

vertex_colors_inv = {
    "tomato": [
        "Commerzbank",
        "ING",
        "HBOS",
        "Lloyds TSB",
        "Royal Bank Scotland",
        "Deutsche Bank",
        "Barclays",
        "Unicredito",
        "Santander",
        "BNP Paribas",
        "Cr. Suisse",
        "IFI",
        "Friends Provident",
        "Soc. Generale",
        "Mediobanca",
        "BNP Paribas",
        "HSBC",
        "UBS",
        "Aberdeen",
        "Intesa-Sanpaolo",
        "Generali",
    ],
    "limegreen": [
        "Nomura",
        "Sumitomomitsui",
        "Sumitomo",
        "Mitsubishi UFJ",
    ],
}
vertex_colors = ["mediumblue" for x in g.vs["name"]]
for color, names in vertex_colors_inv.items():
    for name in names:
        try:
            idx = g.vs.find(name=name).index
        except ValueError:
            continue

        vertex_colors[idx] = color


fig, ax = plt.subplots(figsize=(9, 9))
ipx.network(
    g,
    layout,
    ax=ax,
    edge_color=g.es["color"],
    edge_curved=True,
    edge_tension=0.7,
    vertex_facecolor=vertex_colors,
    vertex_edgecolor="#111",
    vertex_zorder=3,
    vertex_labels=list(g.vs["name"]),
    vertex_label_color="black",
    vertex_label_size=8,
    edge_arrow_marker=")>",
    edge_arrow_width=4,
    edge_arrow_height=7,
    edge_linewidth=g.es["linewidth"],
)

fig.tight_layout()

plt.ion(); plt.show()



"""
Arrow lawn
==========

This example demonstrates the visualisation of a network with 1,000 nodes and a few thousand
edges, using a custom layout that resembles a lawn of arrows.
"""

import numpy as np
import iplotx as ipx
import matplotlib.pyplot as plt

n = 1000  # Number of nodes
k = 3
layout = np.random.rand(n, 2)
nodes = np.arange(n)
edges = []
for i, v1 in enumerate(layout):
    found = 0
    dv = layout - v1
    for j in (dv**2).sum(axis=1).argsort()[1:]:
        angle = 180 / np.pi * np.arctan2(dv[j, 1], dv[j, 0])
        # Only make an edge for neighbors towards the north-east
        if 10 < angle < 80:
            edges.append((i, j))
            found += 1
            if found == k:
                break

# Plot the network using iplotx
fig, ax = plt.subplots(figsize=(7, 7))
ipx.network(
    {'nodes': nodes, 'edges': edges, 'directed': True},
    layout=layout,
    ax=ax,
    vertex_size=5,
    vertex_alpha=0.8,
    vertex_facecolor="seagreen",
    vertex_edgecolor="none",
    edge_arrow_width=3,
    edge_alpha=0.5,
)
fig.tight_layout()


# %%
# If you want to start pushing the scalability of iplotx, you can try with 10,000 nodes and
# ~30,000 edges. This takes ~14 seconds on a modern laptop, of which 5 are spent on building
# the network and 9 are spent inside of ``iplotx`` visualising it:

n = 10000  # Number of nodes
k = 3
layout = np.random.rand(n, 2)
nodes = np.arange(n)
edges = []
for i, v1 in enumerate(layout):
    found = 0
    dv = layout - v1
    for j in (dv**2).sum(axis=1).argsort()[1:]:
        angle = 180 / np.pi * np.arctan2(dv[j, 1], dv[j, 0])
        # Only make an edge for neighbors towards the north-east
        if 10 < angle < 80:
            edges.append((i, j))
            found += 1
            if found == k:
                break

# Plot the network using iplotx
fig, ax = plt.subplots(figsize=(9, 9))
ipx.network(
    {'nodes': nodes, 'edges': edges, 'directed': True},
    layout=layout,
    ax=ax,
    vertex_size=5,
    vertex_alpha=0.8,
    vertex_facecolor="seagreen",
    vertex_edgecolor="none",
    edge_arrow_width=3,
    edge_alpha=0.5,
)
fig.tight_layout()



"""
Cell cycle arrest
=================

This example visualises protein protein interactions involved in the "cell cycle arrest" pathway.

It also shows how to use ``iplotx`` to combine network plots with other matplotlib chart types.
"""

import numpy as np
import pandas as pd
import networkx as nx
import matplotlib.pyplot as plt
import iplotx as ipx


node_data = pd.read_csv("data/cell_cycle_arrest_string_network_coordinates.tsv", sep="\t", index_col=0)
edge_data = pd.read_csv("data/cell_cycle_arrest_string_interactions_short.tsv", sep="\t")

g = nx.Graph()
g.add_edges_from(edge_data.iloc[:, :2].values)

layout = node_data[["x_position", "y_position"]]
node_data["color_ipx"] = [list(map(lambda x: float(x) / 256.0, x[4:-1].split(","))) for x in node_data["color"]]
node_data["label_color"] = ["white" if np.mean(x) < 0.4 else "black" for x in node_data["color_ipx"]]


fig = plt.figure(figsize=(9, 6))
ax = plt.subplot2grid((2, 3), (0, 0), colspan=2, rowspan=2)
ax2 = plt.subplot2grid((2, 3), (0, 2))
ax3 = plt.subplot2grid((2, 3), (1, 2))
ipx.network(
    g,
    layout=layout,
    ax=ax,
    vertex_labels=True,
    vertex_marker="r",
    vertex_size="label",
    vertex_facecolor=node_data["color_ipx"].to_dict(),
    vertex_label_color=node_data["label_color"].to_dict(),
    vertex_edgecolor="k",
    vertex_linewidth=0.5,
    edge_alpha=0.2,
    vertex_zorder=3,

)

degrees = np.bincount(list(dict(g.degree()).values()))
bindegs = np.arange(len(degrees))
ax2.bar(bindegs, degrees, color="gray")
ax2.set_xlabel("Node degree")
ax2.set_ylabel("Number of\nproteins")

pagerank = list(nx.pagerank(g).values())
ax3.ecdf(pagerank, color="black", complementary=True)
ax3.axvline(0.1, ls="--", lw=2, color="tomato")
ax3.set_xlabel("Node PageRank")
ax3.set_ylabel("Cumulative")

fig.tight_layout()



"""
Half arrows
===========

This example demonstrated the use of half arrows for directed graphs.
"""

import matplotlib.pyplot as plt
import iplotx as ipx

g = {
    "edges": [
        ("alice", "bob"),
        ("bob", "alice"),
        ("alice", "jago"),
        ("jago", "alice"),
        ("bob", "jago"),
        ("jago", "bob"),
    ],
    "directed": True,
}

layout = {
    "alice": (0, 0),
    "bob": (1, 0),
    "jago": (0.5, 0.85),
}

fig, ax = plt.subplots(figsize=(3, 2.7))
ipx.network(
    g,
    layout=layout,
    ax=ax,
    edge_arrow_marker="|/",
    edge_paralleloffset=-7,
    edge_arrow_width=15,
)

# %%
# A similar effect can be obtained with the other half arrow, by flipping
# the paralleloffset sign as well. In this case the opposite half arrow is shown:

fig, ax = plt.subplots(figsize=(3, 2.7))
ipx.network(
    g,
    layout=layout,
    ax=ax,
    edge_arrow_marker="|\\",
    edge_paralleloffset=7,
    edge_arrow_width=15,
)

# %%
# If no parallel offset is set, the effect is a little different:

fig, axs = plt.subplots(1, 2, figsize=(6, 2.4))
for ax, marker in zip(axs, ["|/", "|\\"]):
    ipx.network(
        g,
        layout=layout,
        ax=ax,
        edge_arrow_marker=marker,
        edge_paralleloffset=0,
        edge_arrow_width=15,
    )



"""
Big curves
==========

This example showcases the ability of `iplotx` to curve edges at a certain tension setting.
"""

import igraph as ig
import matplotlib.pyplot as plt
import iplotx as ipx


g = ig.Graph(edges=[(0, 1), (0, 0)])
layout = [[0, 0], [1, 0]]

fig, ax = plt.subplots()
ipx.plot(
    g,
    layout=layout,
    ax=ax,
    edge_curved=True,
    edge_tension=10,
    edge_looptension=10,
)



"""
Shortest path
=============

This example from networkx shows how to use edge labels and their rotation.
"""

import networkx as nx
import matplotlib.pyplot as plt
import iplotx as ipx

G = nx.Graph()
G.add_nodes_from(["A", "B", "C", "D", "E", "F", "G", "H"])
G.add_edge("A", "B", weight=4)
G.add_edge("A", "H", weight=8)
G.add_edge("B", "C", weight=8)
G.add_edge("B", "H", weight=11)
G.add_edge("C", "D", weight=7)
G.add_edge("C", "F", weight=4)
G.add_edge("C", "I", weight=2)
G.add_edge("D", "E", weight=9)
G.add_edge("D", "F", weight=14)
G.add_edge("E", "F", weight=10)
G.add_edge("F", "G", weight=2)
G.add_edge("G", "H", weight=1)
G.add_edge("G", "I", weight=6)
G.add_edge("H", "I", weight=7)

# Find the shortest path from node A to node E
path = nx.shortest_path(G, "A", "E", weight="weight")

# Create a list of edges in the shortest path
path_edges = list(zip(path, path[1:]))
nx.set_node_attributes(G, nx.spring_layout(G, seed=37), "pos")
nx.set_edge_attributes(
    G,
    {
        (u, v): {
            "color": (
                "red" if (u, v) in path_edges or tuple(reversed((u, v))) in path_edges else "black"
            ),
            "label": d["weight"],
        }
        for u, v, d in G.edges(data=True)
    },
)

fig, ax = plt.subplots()
ipx.plot(
    G,
    ax=ax,
    layout="pos",
    vertex_labels=True,
    edge_labels=True,
)



"""
Leaf edges and cascades
=======================

This short example demonstrates how to combine leaf edges and cascades.
"""

from Bio import Phylo
from io import StringIO
import iplotx as ipx

# Make a tree from a string in Newick format
tree = next(
    Phylo.NewickIO.parse(
        StringIO(
            "(()(()((()())(()()))))",
        )
    )
)

backgrounds = {
    tree.get_nonterminals()[3]: "turquoise",
    tree.get_terminals()[0]: "tomato",
    tree.get_terminals()[1]: "purple",
}

ipx.plotting.tree(
    tree,
    cascade_facecolor=backgrounds,
    cascade_extend=True,
    leaf_labels={leaf: f"L{i + 1}" for i, leaf in enumerate(tree.get_terminals())},
    edge_color="#111",
    leafedge_linewidth=2,
    leafedge_color="steelblue",
    leafedge_linestyle=":",
)



from typing import Optional, Sequence
from contextlib import nullcontext
import numpy as np
import pandas as pd
import matplotlib as mpl
from mpl_toolkits.mplot3d.axes3d import Axes3D
import matplotlib.pyplot as plt

from .typing import (
    GraphType,
    LayoutType,
    GroupingType,
    TreeType,
)
from .network import NetworkArtist
from .groups import GroupingArtist
from .tree import TreeArtist
from .style import context


def network(
    network: Optional[GraphType] = None,
    layout: Optional[LayoutType] = None,
    grouping: Optional[GroupingType] = None,
    vertex_labels: Optional[list | dict | pd.Series | bool] = None,
    node_labels: Optional[list | dict | pd.Series | bool] = None,
    edge_labels: Optional[Sequence] = None,
    ax: Optional[mpl.axes.Axes] = None,
    style: str | dict | Sequence[str | dict] = (),
    title: Optional[str] = None,
    aspect: Optional[str | float] = None,
    margins: float | tuple[float, float] | tuple[float, float, float] = 0,
    strip_axes: bool = True,
    figsize: Optional[tuple[float, float]] = None,
    **kwargs,
) -> list[mpl.artist.Artist]:
    """Plot this network and/or vertex grouping using the specified layout.

    Parameters:
        network: The network to plot. Can be a networkx or igraph graph.
        layout: The layout to use for plotting. If None, a layout will be looked for in the
            network object and, if none is found, an exception is raised. Defaults to None.
        vertex_labels: The labels for the vertices. If None or False, no vertex labels
            will be drawn. If a list, the labels are taken from the list. If a dict, the keys
            should be the vertex IDs and the values should be the labels. If True (a single
            bool value), the vertex IDs will be used as labels.
        node_labels: Same meaning as vertex_labels. This is an alias to help users who prefer
            the word "node" over "vertex". If both vertex_labels and node_labels are specified,
            "node_labels" overrides "vertex_labels".
        edge_labels: The labels for the edges. If None, no edge labels will be drawn. Defaults
            to None.
        ax: The axis to plot on. If None, a new figure and axis will be created. Defaults to
            None.
        style: Apply this style for the objects to plot. This can be a sequence (e.g. list)
            of styles and they will be applied in order.
        title: If not None, set the axes title to this value.
        aspect: If not None, set the aspect ratio of the axis to this value. In 2D, the most
            common value is 1.0, which proportionates x- and y-axes. In 3D, only string
            values are accepted (see the documentation of Axes.set_aspect).
        margins: How much margin to leave around the plot. A higher value (e.g. 0.1) can be
            used as a quick fix when some vertex shapes reach beyond the plot edge. This is
            a fraction of the data limits, so 0.1 means 10% of the data limits will be left
            as margin. A pair (in 2D) or triplet (in 3D) of floats can also be provided and
            applied to each axis separately.
        strip_axes: If True, remove axis spines and ticks. In 3D, only ticks are removed.
        figsize: If ax is None, a new matplotlib Figure is created. This argument specifies
            the (width, height) dimension of the figure in inches. If ax is not None, this
            argument is ignored. If None, the default matplotlib figure size is used.
        kwargs: Additional arguments are treated as an alternate way to specify style. If
            both "style" and additional **kwargs are provided, they are both applied in that
            order (style, then **kwargs).

    Returns:
        A list of mpl.artist.Artist objects, set as a direct child of the matplotlib Axes.
        The list can have one or two elements, depending on whether you are requesting to
        plot a network, a grouping, or both.
    """
    # Equivalence of node_labels and vertex_labels
    if node_labels is not None:
        vertex_labels = node_labels
    del node_labels

    stylecontext = context(style, **kwargs) if style or kwargs else nullcontext()

    with stylecontext:
        if (network is None) and (grouping is None):
            raise ValueError("At least one of network or grouping must be provided.")

        artists = []
        if network is not None:
            nwkart = NetworkArtist(
                network,
                layout,
                vertex_labels=vertex_labels,
                edge_labels=edge_labels,
                transform=mpl.transforms.IdentityTransform(),
            )
            artists.append(nwkart)
            layout = nwkart.get_layout()
        else:
            nwkart = None

        if grouping is not None:
            grpart = GroupingArtist(
                grouping,
                layout,
                network=network,
            )
            layout = grpart.get_layout()
            artists.append(grpart)
        else:
            grpart = None

        if (nwkart is not None) or (grpart is not None):
            ndim = layout.shape[1]
        else:
            ndim = None

        if ax is None:
            if ndim == 3:
                fig = plt.figure(figsize=figsize)
                ax = fig.add_subplot(111, projection="3d")
            else:
                fig, ax = plt.subplots(figsize=figsize)
                ndim = 2
        else:
            # Check that the expected axis projection is used (3d for 3d layouts)
            if ndim == 3:
                assert isinstance(ax, Axes3D)
            elif ndim == 2:
                # NOTE: technically we probably want it to be cartesian (not polar, etc.)
                # but let's be flexible for now and let that request bubble up from users
                assert not isinstance(ax, Axes3D)

        # This is used in 3D for autoscaling
        had_data = ax.has_data()

        if nwkart is not None:
            # Set the figure, which itself sets the dpi scale for vertices, edges,
            # arrows, etc. Now data limits can be computed correctly
            nwkart.set_offset_transform(ax.transData)
            ax.add_artist(nwkart)
            nwkart.axes = ax

        if grpart is not None:
            grpart.set_transform(ax.transData)
            ax.add_artist(grpart)
            grpart.ax = ax

        if title is not None:
            ax.set_title(title)

        if aspect is not None:
            ax.set_aspect(aspect)

        _postprocess_axes(ax, artists, strip=strip_axes, had_data=had_data)

        if np.isscalar(margins):
            margins = [margins] * ndim
        if (margins[0] != 0) or (margins[1] != 0) or ((len(margins) == 3) and (margins[2] != 0)):
            ax.margins(*margins)

        return artists


# Aliases
plot = network
graph = network


def tree(
    tree: Optional[TreeType] = None,
    layout: str | LayoutType = "horizontal",
    directed: bool | str = False,
    vertex_labels: Optional[list | dict | pd.Series | bool] = None,
    node_labels: Optional[list | dict | pd.Series | bool] = None,
    leaf_labels: Optional[list | dict | pd.Series | bool] = None,
    show_support: bool = False,
    ax: Optional[mpl.axes.Axes] = None,
    style: str | dict | Sequence[str | dict] = "tree",
    title: Optional[str] = None,
    aspect: Optional[str | float] = None,
    margins: float | tuple[float, float] = 0,
    strip_axes: bool = True,
    figsize: Optional[tuple[float, float]] = None,
    **kwargs,
) -> TreeArtist:
    """Plot a tree using the specified layout.

    Parameters:
        tree: The tree to plot. Can be a BioPython.Phylo.Tree object.
        layout: The layout to use for plotting.
        directed: If False, do not draw arrows.
        vertex_labels: The labels for the vertices. If None or False, no vertex labels. Also
            read leaf_labels for leaf nodes.
        node_labels: Same meaning as vertex_labels. This is an alias to help users who prefer
            the word "node" over "vertex". If both vertex_labels and node_labels are specified,
            "node_labels" overrides "vertex_labels".
        leaf_labels: The labels for the leaf nodes. If None or False, no leaf labels are used
            except if vertex_labels are specified for leaf nodes. This argument and the
            previous vertex_labels provide somewhat redundant functionality but have quite
            different default behaviours for distinct use cases. This argument is typically
            useful for labels that are specific to leaf nodes only (e.g. species in a
            phylogenetic tree), whereas vertex_labels is typically used for labels that apply
            to internal nodes too (e.g. branch support values). This redundancy is left on
            purpose to allow for maximal style flexibility.
        show_support: If True, show the support values for the nodes (assumed to be from 0 to 100,
            rounded to nearest integer). If both this parameter and vertex_labels are set,
            show_support takes precedence and hides the vertex labels.
        ax: The axis to plot on. If None, a new figure and axis will be created. Defaults to
            None.
        style: Apply this style for the objects to plot. This can be a sequence (e.g. list)
            of styles and they will be applied in order.
        title: If not None, set the axes title to this value.
        aspect: If not None, set the aspect ratio of the axis to this value. The most common
            value is 1.0, which proportionates x- and y-axes.
        margins: How much margin to leave around the plot. A higher value (e.g. 0.1) can be
            used as a quick fix when some vertex shapes reach beyond the plot edge. This is
            a fraction of the data limits, so 0.1 means 10% of the data limits will be left
            as margin.
        strip_axes: If True, remove axis spines and ticks.
        figsize: If ax is None, a new matplotlib Figure is created. This argument specifies
            the (width, height) dimension of the figure in inches. If ax is not None, this
            argument is ignored. If None, the default matplotlib figure size is used.
        kwargs: Additional arguments are treated as an alternate way to specify style. If
            both "style" and additional **kwargs are provided, they are both applied in that
            order (style, then **kwargs).

    Returns:
        A TreeArtist object, set as a direct child of the matplotlib Axes.
    """
    # Equivalence of node_labels and vertex_labels
    if node_labels is not None:
        vertex_labels = node_labels
    del node_labels

    stylecontext = context(style, **kwargs) if style or kwargs else nullcontext()

    with stylecontext:
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)

        artist = TreeArtist(
            tree=tree,
            layout=layout,
            directed=directed,
            transform=mpl.transforms.IdentityTransform(),
            offset_transform=ax.transData,
            vertex_labels=vertex_labels,
            leaf_labels=leaf_labels,
            show_support=show_support,
        )
        ax.add_artist(artist)
        artist.set_figure(ax.figure)

        if title is not None:
            ax.set_title(title)

        if aspect is not None:
            ax.set_aspect(aspect)

        _postprocess_axes(ax, [artist], strip=strip_axes)

        if np.isscalar(margins):
            margins = (margins, margins)
        if (margins[0] != 0) or (margins[1] != 0):
            ax.margins(*margins)

    return artist


# INTERNAL ROUTINES
def _postprocess_axes(ax, artists, strip=True, had_data=None):
    """Postprocess axis after plotting."""

    if strip:
        if not isinstance(ax, Axes3D):
            # Despine
            ax.spines["right"].set_visible(False)
            ax.spines["top"].set_visible(False)
            ax.spines["left"].set_visible(False)
            ax.spines["bottom"].set_visible(False)

        # Remove axis ticks
        ax.set_xticks([])
        ax.set_yticks([])
        if isinstance(ax, Axes3D):
            ax.set_zticks([])

    # NOTE: bboxes appear to be not that well defined in 3D axes
    # instead, there is a dedicated function that is a little
    # pedestrian
    if isinstance(ax, Axes3D):
        for art in artists:
            XYZ = art.get_layout().values.T
            if ax._zmargin < 0.05 and XYZ[0].size > 0:
                ax.set_zmargin(0.05)
            ax.auto_scale_xyz(
                *XYZ,
                had_data=had_data,
            )
            # NOTE: breaking is not needed, worst case it will
            # autoscale twice (for network and grouping), which
            # is better, at this stage of development, than
            # trying to be too clever by doing the math outselves
    else:
        # Set new data limits
        bboxes = []
        for art in artists:
            bboxes.append(art.get_datalim(ax.transData))
        bbox = mpl.transforms.Bbox.union(bboxes)
        ax.update_datalim(bbox)

        # Autoscale for x/y axis limits
        ax.autoscale_view()



"""
Support module with geometry- and path-related functions for edges.

3D geometry is in a separate module.
"""

from typing import (
    Optional,
    Sequence,
)
from math import atan2, tan, pi
import numpy as np
import matplotlib as mpl

from ..typing import (
    Pair,
)
from .ports import _get_port_unit_vector


def _compute_loops_per_angle(nloops, angles):
    if len(angles) == 0:
        return [(0, 2 * pi, nloops)]

    angles_sorted_closed = list(sorted(angles))
    angles_sorted_closed.append(angles_sorted_closed[0] + 2 * pi)
    deltas = np.diff(angles_sorted_closed)

    # Now we have the deltas and the total number of loops
    # 1. Assign all loops to the largest wedge
    idx_dmax = deltas.argmax()
    if nloops == 1:
        return [
            (
                angles_sorted_closed[idx_dmax],
                angles_sorted_closed[idx_dmax + 1],
                nloops,
            )
        ]

    # 2. Check if any other wedges are larger than this
    # If not, we are done (this is the algo in igraph)
    dsplit = deltas[idx_dmax] / nloops
    if (deltas > dsplit).sum() < 2:
        return [
            (
                angles_sorted_closed[idx_dmax],
                angles_sorted_closed[idx_dmax + 1],
                nloops,
            )
        ]

        # 3. Check how small the second-largest wedge would become
    idx_dsort = np.argsort(deltas)
    return [
        (
            angles_sorted_closed[idx_dmax],
            angles_sorted_closed[idx_dmax + 1],
            nloops - 1,
        ),
        (
            angles_sorted_closed[idx_dsort[-2]],
            angles_sorted_closed[idx_dsort[-2] + 1],
            1,
        ),
    ]


def _get_shorter_edge_coords(vpath, vsize, theta, shrink=0):
    """Get the coordinates of an edge tip such that it touches the vertex border.

    Parameters:
        vpath: the vertex path, in figure coordinates (so scaled by dpi).
        vsize: the vertex max size, in figure coordinates (so scaled by dpi).
        theta: the angle of the edge inpinging into the vertex, in radians, in figure coordinates.
        shrink: additional shrinking of the edge, in figure coordinates (so scaled by dpi).
    """
    # Bound theta from -pi to pi (why is that not guaranteed?)
    theta = (theta + pi) % (2 * pi) - pi

    # Size zero vertices need no shortening
    if vsize == 0:
        ve = np.array([0, 0])
    else:
        for i in range(len(vpath)):
            v1 = vpath.vertices[i]
            v2 = vpath.vertices[(i + 1) % len(vpath)]
            theta1 = atan2(*((v1)[::-1]))
            theta2 = atan2(*((v2)[::-1]))

            # atan2 ranges ]-3.14, 3.14]
            # so it can be that theta1 is -3 and theta2 is +3
            # therefore we need two separate cases, one that cuts at pi and one at 0
            cond1 = theta1 <= theta <= theta2
            cond2 = (
                (theta1 + 2 * pi) % (2 * pi)
                <= (theta + 2 * pi) % (2 * pi)
                <= (theta2 + 2 * pi) % (2 * pi)
            )
            if cond1 or cond2:
                break
        else:
            raise ValueError("Angle for patch not found")

        # The edge meets the patch of the vertex on the v1-v2 size,
        # at angle theta from the center
        mtheta = tan(theta)
        if v2[0] == v1[0]:
            xe = v1[0]
        else:
            m12 = (v2[1] - v1[1]) / (v2[0] - v1[0])
            xe = (v1[1] - m12 * v1[0]) / (mtheta - m12)
        ye = mtheta * xe
        ve = np.array([xe, ye])

    ve = ve * vsize

    # Shrink (assuming dpi scaling is already applied to the shrink)
    ve += shrink * np.array([np.cos(theta), np.sin(theta)])

    return ve


def _fix_parallel_edges_straight(
    paths,
    indices,
    indices_inv,
    trans,
    trans_inv,
    paralleloffset=3,
):
    """Offset parallel edges along the same path."""
    ntot = len(indices) + len(indices_inv)

    # This is straight so two vertices anyway
    # NOTE: all paths will be the same, which is why we need to offset them
    vs, ve = trans(paths[indices[0]].vertices)

    # Move orthogonal to the line
    fracs = (vs - ve) / np.sqrt(((vs - ve) ** 2).sum()) @ np.array([[0, 1], [-1, 0]])

    # NOTE: for now treat both direction the same
    for i, idx in enumerate(indices + indices_inv):
        # Offset the path
        paths[idx].vertices = trans_inv(
            trans(paths[idx].vertices) + fracs * paralleloffset * (i - ntot / 2)
        )


def _compute_loop_path(
    vcoord_fig,
    vpath,
    vsize,
    angle1,
    angle2,
    trans_inv,
    looptension,
    shrink=0,
):
    # Shorten at starting angle
    start = _get_shorter_edge_coords(vpath, vsize, angle1, shrink) + vcoord_fig
    # Shorten at end angle
    end = _get_shorter_edge_coords(vpath, vsize, angle2, shrink) + vcoord_fig

    aux1 = (start - vcoord_fig) * looptension + vcoord_fig
    aux2 = (end - vcoord_fig) * looptension + vcoord_fig

    vertices = np.vstack(
        [
            start,
            aux1,
            aux2,
            end,
        ]
    )
    codes = ["MOVETO"] + ["CURVE4"] * 3

    # Offset to place and transform to data coordinates
    vertices = trans_inv(vertices)
    codes = [getattr(mpl.path.Path, x) for x in codes]
    path = mpl.path.Path(
        vertices,
        codes=codes,
    )
    return path


def _compute_edge_path_straight(
    vcoord_data,
    vpath_fig,
    vsize_fig,
    trans,
    trans_inv,
    layout_coordinate_system: str = "cartesian",
    shrink: float = 0,
    **kwargs,
):
    if layout_coordinate_system not in ("cartesian", "polar"):
        raise ValueError(
            f"Layout coordinate system not supported for straight edges: {layout_coordinate_system}.",
        )

    if layout_coordinate_system == "polar":
        r0, theta0 = vcoord_data[0]
        r1, theta1 = vcoord_data[1]
        vcoord_data_cart = np.array(
            [
                [r0 * np.cos(theta0), r0 * np.sin(theta0)],
                [r1 * np.cos(theta1), r1 * np.sin(theta1)],
            ]
        )
    else:
        vcoord_data_cart = vcoord_data

    # Coordinates in figure (default) coords
    vcoord_fig = trans(vcoord_data_cart)

    points = []

    # Angle of the straight line
    theta = atan2(*((vcoord_fig[1] - vcoord_fig[0])[::-1]))

    # Shorten at starting vertex
    vs = _get_shorter_edge_coords(vpath_fig[0], vsize_fig[0], theta, shrink) + vcoord_fig[0]
    points.append(vs)

    # Shorten at end vertex
    ve = _get_shorter_edge_coords(vpath_fig[1], vsize_fig[1], theta + pi, shrink) + vcoord_fig[1]
    points.append(ve)

    codes = ["MOVETO", "LINETO"]
    path = mpl.path.Path(
        points,
        codes=[getattr(mpl.path.Path, x) for x in codes],
    )
    path.vertices = trans_inv(path.vertices)
    return path, (theta, theta + np.pi)


def _compute_edge_path_waypoints(
    waypoints,
    vcoord_data,
    vpath_fig,
    vsize_fig,
    trans,
    trans_inv,
    layout_coordinate_system: str = "cartesian",
    points_per_curve: int = 30,
    ports: Pair[Optional[str]] = (None, None),
    shrink: float = 0,
    **kwargs,
):
    if not isinstance(waypoints, str):
        # Only cartesian coordinates supported for numerical waypoints for now
        assert layout_coordinate_system == "cartesian"

        waypoints = trans(np.array(waypoints, ndmin=2))

        # Coordinates in figure (default) coords
        vcoord_fig = trans(vcoord_data)

        # Angles of the straight lines
        thetas = [None, None]
        vshorts = [None, None]
        for i in range(2):
            # This picks always the first waypoint for i == 0,
            # the last waypoint for i == 1. They might be the same.
            waypoint = waypoints[-i]
            if ports[i] is None:
                thetas[i] = atan2(*((waypoint - vcoord_fig[i])[::-1]))
            else:
                thetas[i] = atan2(*(_get_port_unit_vector(ports[i], trans_inv)[::-1]))

            # Shorten at vertex border
            vshorts[i] = (
                _get_shorter_edge_coords(vpath_fig[i], vsize_fig[i], thetas[i], shrink)
                + vcoord_fig[i]
            )

        points = [vshorts[0]] + list(waypoints) + [vshorts[1]]
        codes = ["MOVETO"] + ["LINETO"] * len(waypoints) + ["LINETO"]
        angles = tuple(thetas)

    elif waypoints in ("x0y1", "y0x1"):
        assert layout_coordinate_system == "cartesian"

        # Coordinates in figure (default) coords
        vcoord_fig = trans(vcoord_data)

        if waypoints == "x0y1":
            waypoint = np.array([vcoord_fig[0][0], vcoord_fig[1][1]])
        else:
            waypoint = np.array([vcoord_fig[1][0], vcoord_fig[0][1]])

        # Angles of the straight lines
        thetas = [None, None]
        vshorts = [None, None]
        for i in range(2):
            if ports[i] is None:
                thetas[i] = atan2(*((waypoint - vcoord_fig[i])[::-1]))
            else:
                thetas[i] = atan2(*(_get_port_unit_vector(ports[i], trans_inv)[::-1]))

            # Shorten at vertex border
            vshorts[i] = (
                _get_shorter_edge_coords(vpath_fig[i], vsize_fig[i], thetas[i], shrink)
                + vcoord_fig[i]
            )

        # Shorten waypoints to keep the angles right
        if waypoints == "x0y1":
            waypoint[0] = vshorts[0][0]
            waypoint[1] = vshorts[1][1]
        else:
            waypoint[1] = vshorts[0][1]
            waypoint[0] = vshorts[1][0]

        points = [vshorts[0], waypoint, vshorts[1]]
        codes = ["MOVETO", "LINETO", "LINETO"]
        angles = tuple(thetas)
    elif waypoints in ("xmidy0,xmidy1", "x0ymid,x1ymid"):
        # S-shaped orthogonal line
        assert layout_coordinate_system == "cartesian"

        # Coordinates in figure (default) coords
        vcoord_fig = trans(vcoord_data)

        if waypoints == "xmidy0,xmidy1":
            xmid = 0.5 * (vcoord_fig[0][0] + vcoord_fig[1][0])
            waypoint_array = np.array(
                [
                    [xmid, vcoord_fig[0][1]],
                    [xmid, vcoord_fig[1][1]],
                ]
            )
        else:
            ymid = 0.5 * (vcoord_fig[0][1] + vcoord_fig[1][1])
            waypoint_array = np.array(
                [
                    [vcoord_fig[0][0], ymid],
                    [vcoord_fig[1][0], ymid],
                ]
            )

        # Angles of the straight lines
        thetas = []
        vshorts = []
        for i in range(2):
            if ports[i] is None:
                theta = atan2(*((waypoint_array[i] - vcoord_fig[i])[::-1]))
            else:
                theta = atan2(*(_get_port_unit_vector(ports[i], trans_inv)[::-1]))

            # Shorten at vertex border
            vshort = (
                _get_shorter_edge_coords(vpath_fig[i], vsize_fig[i], theta, shrink) + vcoord_fig[i]
            )
            thetas.append(theta)
            vshorts.append(vshort)

        points = [vshorts[0], waypoint_array[0], waypoint_array[1], vshorts[1]]
        codes = ["MOVETO", "LINETO", "LINETO", "LINETO"]
        angles = tuple(thetas)

    elif waypoints == "r0a1":
        assert layout_coordinate_system == "polar"

        r0, alpha0 = vcoord_data[0]
        r1, alpha1 = vcoord_data[1]
        idx_inner = np.argmin([r0, r1])
        idx_outer = 1 - idx_inner
        alpha_outer = [alpha0, alpha1][idx_outer]

        betas = np.linspace(alpha0, alpha1, points_per_curve)
        waypoints = [r0, r1][idx_inner] * np.vstack([np.cos(betas), np.sin(betas)]).T
        endpoint = [r0, r1][idx_outer] * np.array([np.cos(alpha_outer), np.sin(alpha_outer)])
        points = np.array(list(waypoints) + [endpoint])
        points = trans(points)
        codes = ["MOVETO"] + ["LINETO"] * len(waypoints)
        angles = (alpha0 + pi / 2, alpha1)

    else:
        raise NotImplementedError(
            f"Edge shortening with waypoints not implemented yet: {waypoints}.",
        )

    path = mpl.path.Path(
        points,
        codes=[getattr(mpl.path.Path, x) for x in codes],
    )

    path.vertices = trans_inv(path.vertices)
    return path, angles


def _compute_edge_path_curved(
    tension,
    vcoord_data,
    vpath_fig,
    vsize_fig,
    trans,
    trans_inv,
    ports: Pair[Optional[str]] = (None, None),
    shrink: float = 0,
):
    """Shorten the edge path along a cubic Bezier between the vertex centres.

    The most important part is that the derivative of the Bezier at the start
    and end point towards the vertex centres: people notice if they do not.
    """

    # Coordinates in figure (default) coords
    vcoord_fig = trans(vcoord_data)

    dv = vcoord_fig[1] - vcoord_fig[0]
    edge_straight_length = np.sqrt((dv**2).sum())

    auxs = [None, None]
    for i in range(2):
        if ports[i] is not None:
            der = _get_port_unit_vector(ports[i], trans_inv)
            auxs[i] = der * edge_straight_length * tension + vcoord_fig[i]

    # Both ports defined, just use them and hope for the best
    # Obviously, if the user specifies ports that make no sense,
    # this is going to be a (technically valid) mess.
    if all(aux is not None for aux in auxs):
        pass

    # If no ports are specified (the most common case), compute
    # the Bezier and shorten it
    elif all(aux is None for aux in auxs):
        # Put auxs along the way
        auxs = np.array(
            [
                vcoord_fig[0] + 0.33 * dv,
                vcoord_fig[1] - 0.33 * dv,
            ]
        )
        # Right rotation from the straight edge
        dv_rot = -0.1 * dv @ np.array([[0, 1], [-1, 0]])
        # Shift the auxs orthogonal to the straight edge
        auxs += dv_rot * tension

    # First port is defined
    elif (auxs[0] is not None) and (auxs[1] is None):
        auxs[1] = auxs[0]

    # Second port is defined
    else:
        auxs[0] = auxs[1]

    vs = [None, None]
    thetas = [None, None]
    for i in range(2):
        thetas[i] = atan2(*((auxs[i] - vcoord_fig[i])[::-1]))
        vs[i] = (
            _get_shorter_edge_coords(vpath_fig[i], vsize_fig[i], thetas[i], shrink) + vcoord_fig[i]
        )

    path = {
        "vertices": [
            vs[0],
            auxs[0],
            auxs[1],
            vs[1],
        ],
        "codes": ["MOVETO"] + ["CURVE4"] * 3,
    }

    path = mpl.path.Path(
        path["vertices"],
        codes=[getattr(mpl.path.Path, x) for x in path["codes"]],
    )

    # Return to data transform
    path.vertices = trans_inv(path.vertices)
    return path, tuple(thetas)


def _compute_edge_path(
    *args,
    tension: float = 0,
    waypoints: str | tuple[float, float] | Sequence[tuple[float, float]] | np.ndarray = "none",
    ports: Pair[Optional[str]] = (None, None),
    layout_coordinate_system: str = "cartesian",
    **kwargs,
):
    """Compute the edge path in a few different ways."""
    if (waypoints != "none") and (tension != 0):
        raise ValueError("Waypoints not supported for curved edges.")

    if waypoints != "none":
        return _compute_edge_path_waypoints(
            waypoints,
            *args,
            layout_coordinate_system=layout_coordinate_system,
            ports=ports,
            **kwargs,
        )

    if tension == 0:
        return _compute_edge_path_straight(
            *args,
            layout_coordinate_system=layout_coordinate_system,
            **kwargs,
        )

    return _compute_edge_path_curved(
        tension,
        *args,
        ports=ports,
        **kwargs,
    )



import numpy as np
import pandas as pd
import matplotlib as mpl
from matplotlib import patches
import iplotx as ipx


def vertex_collection(**kwargs):
    if "layout" not in kwargs:
        kwargs["layout"] = pd.DataFrame(
            data=np.array([[0, 0], [1, 1]], np.float64),
            index=["A", "B"],
        )
    return ipx.artists.VertexCollection(**kwargs)


def edge_collection(**kwargs):
    if vertex_collection not in kwargs:
        kwargs["vertex_collection"] = vertex_collection()
    if "vertex_ids" not in kwargs:
        vids = kwargs["vertex_collection"].get_index()
        nedges = 4
        kwargs["vertex_ids"] = [
            [vids[i % len(vids)], vids[(i + 1) % len(vids)]] for i in range(nedges)
        ]
        kwargs["patches"] = [
            patches.PathPatch(mpl.path.Path(np.array([[0, 0], [1, 1]]))),
        ] * len(kwargs["vertex_ids"])

    return ipx.artists.EdgeCollection(**kwargs)


def test_hotload():
    edges = edge_collection()
    res = ipx.artists.NetworkArtist.from_edgecollection(
        edges,
    )
    assert isinstance(res, ipx.artists.NetworkArtist)

    res2 = ipx.artists.NetworkArtist.from_other(res)
    assert isinstance(res2, ipx.artists.NetworkArtist)



"""
Edge waypoints
==============

This example demonstrates the use of edge waypoints.
"""

import matplotlib.pyplot as plt
import numpy as np
import iplotx as ipx

g = {
    "edges": [
        ("A", "B"),
        ("B", "C"),
        ("B", "D"),
        ("A", "E"),
    ],
}
layout = {
    "A": (0, 0),
    "B": (0, 1),
    "C": (1, 2),
    "D": (0, 2),
    "E": (-1, 1),
}

fig, ax = plt.subplots()
ipx.network(
    g,
    layout=layout,
    ax=ax,
    vertex_labels=True,
    edge_waypoints=["none", [(1, 1), (0.5, 1.5), (1, 1.5)], "none", "none"],
)
plt.ion(); plt.show()

# %%
# .. note::
#   Edge waypoints are not currently compatible with curved edges.



"""
Main style module for iplotx.
"""

from typing import (
    Any,
    Iterable,
    Optional,
    Sequence,
)
from collections.abc import Hashable
from contextlib import contextmanager
import numpy as np
import pandas as pd

from ..utils.style import (
    copy_with_deep_values,
    sanitize_leaves,
    update_style,
    sanitize_ambiguous,
)
from .library import style_library
from .leaf_info import (
    style_leaves,
    nonrotating_leaves,
)


# Prepopulate default style, it's used later as a backbone for everything else
default = style_library["default"]
styles = {
    "default": default,
}


current = copy_with_deep_values(styles["default"])


def get_style(name: str = "", *args) -> dict[str, Any]:
    """Get a *deep copy* of the chosen style.

    Parameters:
        name: The name of the style to get. If empty, the current style is returned.
            Substyles can be obtained by using a dot notation, e.g. "default.vertex".
            If "name" starts with a dot, it means a substyle of the current style.
        *args: A single argument is accepted. If present, this value (usually a
            dictionary) is returned if the queried style is not found. For example,
            get_style(".nonexistent") raises an Exception but
            get_style("nonexistent", {}) does not, returning an empty dict instead.
    Returns:
        The requected style or substyle.

    NOTE: The deep copy is a little different from standard deep copies. Here, keys
        (which need to be hashables) are never copied, but values are. This can be
        useful for hashables that change hash upon copying, such as Biopython's
        tree nodes.
    """
    if len(args) > 1:
        raise ValueError("get_style() accepts at most one additional argument.")

    namelist = name.split(".")
    style = styles
    for i, namei in enumerate(namelist):
        if (i == 0) and (namei == ""):
            style = current
        else:
            if namei in style:
                style = style[namei]
            # NOTE: if asking for a nonexistent, non-leaf style
            # give the benefit of the doubt and set an empty dict
            # which will not fail unless the user tries to enter it
            elif namei not in style_leaves:
                style = {}
            elif len(args) > 0:
                return args[0]
            else:
                raise KeyError(f"Style not found: {name}")

    style = copy_with_deep_values(style)
    return style


def merge_styles(
    styles: Sequence[str | dict[str, Any]] | Iterable[str | dict[str, Any]],
) -> dict[str, Any]:
    """Merge a sequence of styles into a single one.

    Parameters:
        styles: Sequence (list, tuple, etc.) of styles, each either the name of an internal
            style or a dict-like with custom properties.
    Returns:
        The composite style as a dict.
    """

    merged = {}
    for style in styles:
        if isinstance(style, str):
            style = get_style(style)
        else:
            sanitize_leaves(style)
            unflatten_style(style)
            sanitize_ambiguous(style)
        update_style(style, merged)

    return merged


# The following is inspired by matplotlib's style library
# https://github.com/matplotlib/matplotlib/blob/v3.10.3/lib/matplotlib/style/core.py#L45
def use(
    style: Optional[
        str | dict[str, Any] | Sequence[str | dict[str, Any]] | Iterable[str | dict[str, Any]]
    ] = None,
    **kwargs,
):
    """Use iplotx style setting for a style specification.

    The style name of 'default' is reserved for reverting back to
    the default style settings.

    Parameters:
        style: A style specification, currently either a name of an existing style
            or a dict with specific parts of the style to override. The string
            "default" resets the style to the default one. If this is a sequence,
            each style is applied in order.
        **kwargs: Additional style changes to be applied at the end of any style.
    """
    global current

    styles = []
    if isinstance(style, (dict, str)):
        styles.append(style)
    elif style is not None:
        styles.extend(list(style))
    if kwargs:
        styles.append(kwargs)

    # Discard empty styles for speed
    styles = [style for style in styles if style]

    if len(styles) == 0:
        return

    # If the first style is a string (internal style), apply it cold. If it's a
    # dict, apply it hot (on top of the current style(. Any style after the first
    # is always applied hot - otherwise it would invalidate previous styles.
    if not isinstance(styles[0], str):
        # hot insertion on top of current
        styles.insert(0, current)

    old_style = copy_with_deep_values(current)
    try:
        current = merge_styles(styles)
    except:
        current = old_style
        raise


def reset() -> None:
    """Reset to default style."""
    global current
    current = copy_with_deep_values(styles["default"])


@contextmanager
def context(
    style: Optional[
        str | dict[str, Any] | Sequence[str | dict[str, Any]] | Iterable[str | dict[str, Any]]
    ] = None,
    **kwargs,
):
    """Create a style context for iplotx.

    Parameters:
        style: A single style specification or a list of style specifications, which are then
            applied in order. Each style can be a string (for an existing style) or a dictionary
            with the elements that are to change.
        **kwargs: Additional style changes to be applied at the end of all styles.

    Yields:
        A context manager that applies the style and reverts it back to the previous one upon exit.
    """
    current = get_style()
    try:
        use(style, **kwargs)
        yield
    finally:
        use(["default", current])


def unflatten_style(
    style_flat: dict[str, str | dict | int | float],
) -> None:
    """Convert a flat or semi-flat style into a fully structured dict.

    Parameters:
        style_flat: A flat dictionary where keys may contain underscores, which are taken to signify
            subdictionaries.

    NOTE: The dict is changed *in place*.

    Example:
        >>> style = {'vertex_size': 20}
        >>> unflatten_style(style)
        >>> print(style)
        {'vertex': {'size': 20}}
    """

    def _inner(style_flat: dict):
        keys = list(style_flat.keys())

        for key in keys:
            if "_" not in key:
                continue

            keyhead, keytail = key.split("_", 1)
            value = style_flat.pop(key)
            if keyhead not in style_flat:
                style_flat[keyhead] = {
                    keytail: value,
                }
            else:
                style_flat[keyhead][keytail] = value

        for key, value in style_flat.items():
            if isinstance(value, dict) and (key not in style_leaves):
                _inner(value)

    # top-level adjustments
    if "zorder" in style_flat:
        style_flat["network_zorder"] = style_flat["grouping_zorder"] = style_flat.pop("zorder")

    # Begin recursion
    _inner(style_flat)


def rotate_style(
    style: dict[str, Any],
    index: Optional[int] = None,
    key: Optional[Hashable] = None,
    key2: Optional[Hashable] = None,
    props: Optional[Sequence[str]] = None,
) -> dict[str, Any]:
    """Rotate leaves of a style for a certain index or key.

    Parameters:
        style: The style to rotate.
        index: The integer to rotate the style leaves into.
        key: For dict-like leaves (e.g. vertex properties specified as a dict-like object over the
            vertices themselves), the key to use for rotation (e.g. the vertex itself).
        key2: For dict-like leaves, a backup key in case the first key fails. If this is None
            or also a failure (i.e. KeyError), default to the empty type constructor for the
            first value of the dict-like style leaf.
        props: The properties to rotate, usually all leaf properties.

    Returns:
        A style with rotated leaves, which describes the properties of a single element (e.g.
        vertex).

    Example:
        >>> style = {'vertex': {'size': [10, 20]}}
        >>> rotate_style(style, index=0)
        {'vertex': {'size': 10}}
    """
    if (index is None) and (key is None):
        raise ValueError("At least one of 'index' or 'key' must be provided to rotate_style.")

    if props is None:
        props = tuple(prop for prop in style_leaves if prop not in nonrotating_leaves)

    style = copy_with_deep_values(style)

    for prop in props:
        val = style.get(prop, None)
        if val is None:
            continue
        # Try integer indexing for ordered types
        if (index is not None) and isinstance(val, (tuple, list, np.ndarray, pd.Index, pd.Series)):
            # NOTE: cannot cast to ndarray because rotation might involve
            # cross-type lists (e.g. ["none", False])
            style[prop] = list(val)[index % len(val)]
        # Try key indexing for unordered, dict-like types
        if (
            (key is not None)
            and (not isinstance(val, (str, tuple, list, np.ndarray)))
            and hasattr(val, "__getitem__")
        ):
            # If only a subset of keys is provided, try the default value a la
            # defaultdict. If that fails, use an empty constructor
            if key in val:
                newval = val[key]
            elif key2 is not None and (key2 in val):
                newval = val[key2]
            else:
                try:
                    newval = val[key]
                except KeyError:
                    valtype = type(next(iter(val.values())))
                    newval = valtype()
            style[prop] = newval

    return style


def add_style(name: str, style: dict[str, Any]) -> None:
    """Add a style to the default dictionary of styles.

    Parameters:
        name: The name of the style to add.
        style: A dictionary with the style properties to add.
    """
    with context(["default", style]):
        styles[name] = get_style()


# Populate style library (default is already there)
for name, style in style_library.items():
    if name != "default":
        add_style(name, style)
        del name, style



import unittest
import pytest
import numpy as np
import pandas as pd
import matplotlib as mpl

mpl.use("agg")
import matplotlib.pyplot as plt
import iplotx as ipx

from utils import image_comparison


class GraphTestRunner(unittest.TestCase):
    @property
    def layout_small_ring(self):
        coords = [
            [1.015318095035966, 0.03435580194714975],
            [0.29010409851547664, 1.0184451153265959],
            [-0.8699239050738742, 0.6328259400443561],
            [-0.8616466426732888, -0.5895891303732176],
            [0.30349699041342515, -0.9594640169691343],
        ]
        return coords

    def make_ring(self):
        g = {
            "edges": [
                (0, 1),
                (1, 2),
                (2, 3),
                (3, 4),
                (4, 0),
            ]
        }
        return g

    def test_with_nodes(self):
        g = self.make_ring()
        g["nodes"] = [0, 1, 2, 3, 4]
        ipx.ingest.ingest_network_data(
            g,
        )

    def test_with_False_labels(self):
        g = self.make_ring()
        ipx.ingest.ingest_network_data(
            g,
            vertex_labels=False,
        )

    def test_with_True_labels(self):
        g = self.make_ring()
        ipx.ingest.ingest_network_data(
            g,
            vertex_labels=True,
        )

    def test_with_invalid_labels(self):
        g = self.make_ring()
        with pytest.raises(ValueError):
            ipx.ingest.ingest_network_data(
                g,
                vertex_labels=["x", "y"],
            )

    def test_with_no_edges(self):
        g = {
            "nodes": [0, 1],
        }
        ipx.ingest.ingest_network_data(
            g,
        )

    def test_with_layout_dataframe(self):
        g = self.make_ring()
        ipx.ingest.ingest_network_data(
            g,
            layout=pd.DataFrame(
                np.zeros((5, 2), np.float64),
            ),
        )

    def test_with_layout_dict(self):
        g = self.make_ring()
        ipx.ingest.ingest_network_data(
            g,
            layout=pd.DataFrame(
                np.zeros((5, 2), np.float64),
            ).T.to_dict(),
        )

    def test_count_nodes(self):
        g = self.make_ring()
        network_data = ipx.ingest.ingest_network_data(g)
        ptype = network_data["network_library"]
        provider = ipx.ingest.data_providers["network"][ptype]
        nvertices = provider(g).number_of_vertices()
        assert nvertices == 5

        g["nodes"] = [0, 1, 2, 3, 4]
        nvertices = provider(g).number_of_vertices()
        assert nvertices == 5

    @image_comparison(baseline_images=["graph_basic"], remove_text=True)
    def test_basic(self):
        g = self.make_ring()
        fig, ax = plt.subplots(figsize=(3, 3))
        ipx.plot(g, ax=ax, layout=self.layout_small_ring)

    @image_comparison(baseline_images=["graph_directed"], remove_text=True)
    def test_directed(self):
        g = self.make_ring()
        g["directed"] = True
        fig, ax = plt.subplots(figsize=(3, 3))
        ipx.plot(g, ax=ax, layout=self.layout_small_ring)

    @image_comparison(baseline_images=["graph_labels"], remove_text=True)
    def test_labels(self):
        g = self.make_ring()
        fig, ax = plt.subplots(figsize=(3, 3))
        ipx.plot(
            network=g,
            ax=ax,
            layout=self.layout_small_ring,
            vertex_labels=["1", "2", "3", "4", "5"],
            style={
                "vertex": {
                    "size": 20,
                    "label": {
                        "color": "white",
                        "size": 10,
                    },
                }
            },
        )



"""
Style library
=============

This example shows the same network plotted with a variery of styles from the
`iplotx` style library.

.. tip::
  You can further modify any of these styles with a custom dictionary (see :doc:`plot_style`).
"""

import igraph as ig
import matplotlib.pyplot as plt
import iplotx as ipx

styles = [
    "default",
    "hollow",
    "tree",
    "greyscale",
    "rededge",
    "rednode",
    "networkx",
    "igraph",
    "eerie",
    "unicorn",
]

g = ig.Graph.Ring(9, directed=True)
layout = g.layout("circle")

fig, axs = plt.subplots(5, 2, figsize=(7.5, 20))
for style, ax in zip(styles, axs.ravel()):
    ipx.plot(
        g,
        ax=ax,
        layout=layout,
        style=style,
        vertex_labels=style in ("default", "hollow", "unicorn", "networkx", "igraph"),
        title=style,
        margins=0.1 if style == "unicorn" else 0,
    )

# %%
# For completeness, below the same plot but for the undirected ring graph:

g = ig.Graph.Ring(9, directed=False)

fig, axs = plt.subplots(5, 2, figsize=(7.5, 20))
for style, ax in zip(styles, axs.ravel()):
    ipx.plot(
        g,
        ax=ax,
        layout="circle",
        style=style,
        vertex_labels=style in ("default", "hollow", "unicorn", "networkx", "igraph"),
        title=style,
        margins=0.1 if style == "unicorn" else 0,
    )



from typing import (
    Optional,
    Sequence,
)
from collections.abc import Hashable
import numpy as np
import pandas as pd

from ....typing import (
    LayoutType,
)
from ...typing import (
    NetworkDataProvider,
    NetworkData,
)
from ....utils.internal import (
    _make_layout_columns,
)


class SimpleNetworkDataProvider(NetworkDataProvider):
    def __call__(
        self,
        layout: Optional[LayoutType] = None,
        vertex_labels: Optional[Sequence[str] | dict[Hashable, str] | pd.Series] = None,
        edge_labels: Optional[Sequence[str] | dict[str]] = None,
    ) -> NetworkData:
        """Create network data object for iplotx from a simple Python object."""
        network = self.network
        directed = self.is_directed()

        # Recast vertex_labels=False as vertex_labels=None
        if np.isscalar(vertex_labels) and (not vertex_labels):
            vertex_labels = None

        # Vertices are ordered integers, no gaps
        for key in ["nodes", "vertices"]:
            if key in network:
                vertices = network[key]
                break
        else:
            # Infer from edge adjacent vertices, singletons will be missed
            vertices = set()
            for edge in self.network.get("edges", []):
                vertices.add(edge[0])
                vertices.add(edge[1])
        vertices = list(vertices)

        # NOTE: This is underpowered, but it's ok for a simple educational provider
        if isinstance(layout, pd.DataFrame):
            vertex_df = layout.loc[vertices].copy()
        elif isinstance(layout, dict):
            vertex_df = pd.DataFrame(layout).T.loc[vertices]
        else:
            vertex_df = pd.DataFrame(
                index=vertices,
                data=layout,
            )
        ndim = vertex_df.shape[1]
        vertex_df.columns = _make_layout_columns(ndim)

        # Vertex labels
        if vertex_labels is not None:
            if np.isscalar(vertex_labels):
                vertex_df["label"] = vertex_df.index.astype(str)
            elif len(vertex_labels) != len(vertex_df):
                raise ValueError("Vertex labels must be the same length as the number of vertices.")
            else:
                vertex_df["label"] = vertex_labels

        # Edges are a list of tuples, because of multiedges
        tmp = []
        for edge in network.get("edges", []):
            row = {"_ipx_source": edge[0], "_ipx_target": edge[1]}
            tmp.append(row)
        if len(tmp):
            edge_df = pd.DataFrame(tmp)
        else:
            edge_df = pd.DataFrame(columns=["_ipx_source", "_ipx_target"])
        del tmp

        # Edge labels
        if edge_labels is not None:
            edge_df["label"] = edge_labels

        network_data = {
            "vertex_df": vertex_df,
            "edge_df": edge_df,
            "directed": directed,
            "ndim": ndim,
        }
        return network_data

    @staticmethod
    def check_dependencies() -> bool:
        """Check dependencies. Returns True since this provider has no dependencies."""
        return True

    @staticmethod
    def graph_type():
        return dict

    def is_directed(self):
        """Whether the network is directed."""
        return self.network.get("directed", False)

    def number_of_vertices(self):
        """The number of vertices/nodes in the network."""
        for key in ("nodes", "vertices"):
            if key in self.network:
                return len(self.network[key])

        # Default to unique edge adjacent nodes (this will ignore singletons)
        nodes = set()
        for edge in self.network.get("edges", []):
            nodes.add(edge[0])
            nodes.add(edge[1])
        return len(nodes)



from io import StringIO
import pytest
import numpy as np
import pandas as pd
import matplotlib as mpl

mpl.use("agg")
import matplotlib.pyplot as plt
import iplotx as ipx

from utils import image_comparison

Bio = pytest.importorskip("Bio")
from Bio import Phylo  # noqa: E402


@pytest.fixture
def tree():
    tree = next(
        Phylo.NewickIO.parse(
            StringIO(
                "(()(()((()())(()()))))",
            )
        )
    )
    return tree


def test_get_layout(tree):
    art = ipx.artists.TreeArtist(
        tree,
    )
    assert isinstance(art.get_layout(), pd.DataFrame)


def test_get_edge_layout(tree):
    art = ipx.artists.TreeArtist(
        tree,
    )
    with pytest.raises(KeyError):
        art.get_layout(kind="edge")


def test_get_invalid_layout(tree):
    art = ipx.artists.TreeArtist(
        tree,
    )
    with pytest.raises(ValueError):
        art.get_layout(kind="invalid")


def test_get_zero_layout(tree):
    art = ipx.artists.TreeArtist(
        tree,
    )
    art._ipx_internal_data["vertex_df"] = art._ipx_internal_data["vertex_df"][:0]
    bbox = art.get_datalim(None)
    assert isinstance(bbox, mpl.transforms.Bbox)


def test_get_elements(tree):
    art = ipx.artists.TreeArtist(
        tree,
        leaf_labels={leaf: str(i + 1) for i, leaf in enumerate(tree.get_terminals())},
    )
    assert isinstance(art.get_vertices(), ipx.artists.VertexCollection)
    assert isinstance(art.get_edges(), ipx.artists.EdgeCollection)
    assert isinstance(art.get_leaf_vertices(), ipx.artists.VertexCollection)
    assert isinstance(art.get_leaf_edges(), ipx.artists.EdgeCollection)
    assert art.get_leaf_edge_labels() is None

    with ipx.style.context(
        [
            "tree",
            {"leaf": {"deep": False}},
        ]
    ):
        art = ipx.artists.TreeArtist(
            tree,
        )
        assert art.get_leaf_edges() is None
        assert art.get_leaf_edge_labels() is None

    assert art.get_vertex_labels() is None


def test_leaf_labels_array(tree):
    for leaf in tree.get_terminals():
        leaf.name = "hello"
    ipx.artists.TreeArtist(
        tree,
        leaf_labels=True,
    )


@pytest.mark.parametrize(
    "layout,orientation",
    [
        ["horizontal", "right"],
        ["horizontal", "left"],
        ["vertical", "descending"],
        ["vertical", "ascending"],
        ["radial", "clockwise"],
        ["radial", "counterclockwise"],
    ],
)
def test_vertical_leaf_orientations(tree, layout, orientation):
    with ipx.style.context(["tree", {"layout_orientation": orientation}]):
        ipx.artists.TreeArtist(
            tree,
            layout=layout,
            leaf_labels={leaf: str(i + 1) for i, leaf in enumerate(tree.get_terminals())},
        )


def test_leaf_vertices_invalid_orientation(tree):
    with ipx.style.context(["tree", {"layout_orientation": "unsupported"}]):
        with pytest.raises(ValueError):
            ipx.artists.TreeArtist(
                tree,
                layout="vertical",
                leaf_labels={leaf: str(i + 1) for i, leaf in enumerate(tree.get_terminals())},
            )


def test_get_maxdepth_leaf_labels(tree):
    art = ipx.artists.TreeArtist(
        tree,
        leaf_labels={leaf: str(i + 1) for i, leaf in enumerate(tree.get_terminals())},
    )
    maxdepth = art._get_maxdepth_leaf_labels()
    assert isinstance(maxdepth, float)


def test_style_subtree(tree):
    art = ipx.artists.TreeArtist(
        tree,
        leaf_labels={leaf: str(i + 1) for i, leaf in enumerate(tree.get_terminals())},
    )
    art.style_subtree(
        tree.clade[0, 1],
        {"edge": {"color": "red"}},
    )


def test_get_edge_labels(tree):
    art = ipx.artists.TreeArtist(
        tree,
        edge_labels=["A", "B", "C", "D", "E", "F"],
    )
    np.testing.assert_array_equal(
        art.get_edge_labels().get_texts()[:6],
        np.array(["A", "B", "C", "D", "E", "F"]),
    )
    np.testing.assert_array_equal(
        art.get_edge_labels().get_text()[:6],
        np.array(["A", "B", "C", "D", "E", "F"]),
    )


@image_comparison(baseline_images=["tree_basic"], remove_text=True)
def test_basic(tree):
    fig, ax = plt.subplots(figsize=(3, 3))
    ipx.plotting.tree(
        tree=tree,
        ax=ax,
        layout="horizontal",
    )


@image_comparison(baseline_images=["tree_radial"], remove_text=True)
def test_radial(tree):
    fig, ax = plt.subplots(figsize=(3, 3))
    ipx.plotting.tree(
        tree=tree,
        ax=ax,
        layout="radial",
        aspect=1,
    )


@image_comparison(baseline_images=["leaf_labels"], remove_text=True)
def test_leaf_labels(tree):
    leaf_labels = {leaf: str(i + 1) for i, leaf in enumerate(tree.get_terminals())}

    fig, ax = plt.subplots(figsize=(4, 4))
    ipx.plotting.tree(
        tree=tree,
        ax=ax,
        layout="horizontal",
        vertex_labels=leaf_labels,
        margins=0.1,
    )


@image_comparison(baseline_images=["leaf_labels_hmargin"], remove_text=True)
def test_leaf_labels_hmargin(tree):
    leaf_labels = {leaf: str(i + 1) for i, leaf in enumerate(tree.get_terminals())}
    vertex_label_hmargin = {key: [10, 22][(int(x) - 1) % 2] for key, x in leaf_labels.items()}

    fig, ax = plt.subplots(figsize=(4, 4))
    ipx.plotting.tree(
        tree=tree,
        ax=ax,
        layout="horizontal",
        vertex_labels=leaf_labels,
        vertex_label_hmargin=vertex_label_hmargin,
        margins=(0.15, 0),
    )


@image_comparison(baseline_images=["show_support"], remove_text=True)
def test_show_support(tree):
    for node in tree.get_nonterminals():
        if node != tree.root:
            node.confidence = 0.9

    fig, ax = plt.subplots(figsize=(4, 4))
    ipx.plotting.tree(
        tree=tree,
        ax=ax,
        layout="horizontal",
        show_support=True,
    )


@image_comparison(baseline_images=["cascades"], remove_text=True)
def test_cascades(tree):
    backgrounds = {
        tree.get_nonterminals()[3]: "turquoise",
        tree.get_terminals()[0]: "tomato",
        tree.get_terminals()[1]: "purple",
    }

    ipx.plotting.tree(
        tree,
        cascade_facecolor=backgrounds,
        cascade_edgecolor="none",
    )


@image_comparison(baseline_images=["leafedges"], remove_text=True)
def test_leafedges(tree):
    ipx.plotting.tree(
        tree,
        leaf_labels={leaf: str(i + 1) for i, leaf in enumerate(tree.get_terminals())},
    )


@image_comparison(baseline_images=["directed_child"], remove_text=True)
def test_directed_child(tree):
    ipx.plotting.tree(
        tree,
        directed=True,
    )



from typing import (
    Optional,
    Sequence,
)
from collections.abc import Hashable
import importlib
import numpy as np
import pandas as pd

from ....typing import (
    LayoutType,
)
from ...heuristics import (
    normalise_layout,
)
from ...typing import (
    NetworkDataProvider,
    NetworkData,
)
from ....utils.internal import (
    _make_layout_columns,
)


class IGraphDataProvider(NetworkDataProvider):
    def __call__(
        self,
        layout: Optional[LayoutType] = None,
        vertex_labels: Optional[Sequence[str] | dict[Hashable, str] | pd.Series] = None,
        edge_labels: Optional[Sequence[str] | dict[str]] = None,
    ) -> NetworkData:
        """Create network data object for iplotx from an igraph object."""

        # Get layout
        vertex_df = normalise_layout(
            layout,
            network=self.network,
            nvertices=self.number_of_vertices(),
        )
        ndim = vertex_df.shape[1]
        vertex_df.columns = _make_layout_columns(ndim)

        # Vertices are ordered integers, no gaps

        # Vertex labels
        # Recast vertex_labels=False as vertex_labels=None
        if np.isscalar(vertex_labels) and (not vertex_labels):
            vertex_labels = None
        if vertex_labels is not None:
            if np.isscalar(vertex_labels):
                vertex_df["label"] = vertex_df.index.astype(str)
            elif len(vertex_labels) != len(vertex_df):
                raise ValueError("Vertex labels must be the same length as the number of vertices.")
            else:
                vertex_df["label"] = vertex_labels

        # Edges are a list of tuples, because of multiedges
        tmp = []
        for edge in self.network.es:
            row = {"_ipx_source": edge.source, "_ipx_target": edge.target}
            row.update(edge.attributes())
            tmp.append(row)
        if len(tmp):
            edge_df = pd.DataFrame(tmp)
        else:
            edge_df = pd.DataFrame(columns=["_ipx_source", "_ipx_target"])
        del tmp

        # Edge labels
        if edge_labels is not None:
            if len(edge_labels) != len(edge_df):
                raise ValueError("Edge labels must be the same length as the number of edges.")
            edge_df["label"] = edge_labels

        network_data = {
            "vertex_df": vertex_df,
            "edge_df": edge_df,
            "directed": self.is_directed(),
            "ndim": ndim,
        }
        return network_data

    @staticmethod
    def check_dependencies() -> bool:
        return importlib.util.find_spec("igraph") is not None

    @staticmethod
    def graph_type():
        import igraph as ig

        return ig.Graph

    def is_directed(self):
        """Whether the network is directed."""
        return self.network.is_directed()

    def number_of_vertices(self):
        """The number of vertices/nodes in the network."""
        return self.network.vcount()



import copy
from collections import defaultdict

from ..style.leaf_info import (
    style_leaves,
)

try:
    import networkx as nx
except ImportError:
    nx = None


def copy_with_deep_values(style):
    """Make a deep copy of the style dict but do not create copies of the keys."""
    # Defaultdict should be respected
    if hasattr(style, "default_factory"):
        newdict = defaultdict(lambda: style.default_factory())
    else:
        newdict = {}
    for key, value in style.items():
        if isinstance(value, dict):
            newdict[key] = copy_with_deep_values(value)
        else:
            newdict[key] = copy.copy(value)
    return newdict


def sanitize_leaves(style: dict):
    """Sanitize the leaves of a style dictionary.

    Parameters:
        style (dict): A style dictionary.
    Returns:
        None: The style dictionary is modified in place.
    """
    for key, value in style.items():
        if key in style_leaves:
            # Networkx has a few lazy data structures
            # TODO: move this code to provider
            if nx is not None:
                if isinstance(value, nx.classes.reportviews.NodeView):
                    style[key] = dict(value)
                elif isinstance(value, nx.classes.reportviews.EdgeViewABC):
                    style[key] = [v for *e, v in value]

        elif isinstance(value, dict):
            sanitize_leaves(value)


def update_style(new_style: dict, current_style: dict):
    """Update the current style with a new style.

    Parameters:
        new_style (dict): A new style dictionary.
        current_style (dict): The current style dictionary.
    Returns:
        None: The current style dictionary is modified in place.
    """
    for key, value in new_style.items():
        if key not in current_style:
            current_style[key] = value
            continue

        # Style non-leaves are either recurred into or deleted
        if key not in style_leaves:
            if isinstance(value, dict):
                update_style(value, current_style[key])
            elif value is None:
                del current_style[key]
            else:
                raise ValueError(
                    f"Setting non-leaf style value to a non-dict: {key}, {value}",
                )
        else:
            # Style leaves could be incomplete, ensure a sensible default
            if value is None:
                del current_style[key]
                continue

            if not isinstance(value, dict):
                current_style[key] = value
                continue

            if hasattr(value, "default_factory"):
                current_style[key] = value
                continue

            if hasattr(current_style[key], "default_factory"):
                default_value = current_style[key].default_factory()
            else:
                default_value = current_style[key]
            current_style[key] = defaultdict(
                lambda: default_value,
                value,
            )


def sanitize_ambiguous(style: dict):
    """Fix a few ambiguous cases in the style dict.

    Parameters:
        style (dict): A style dictionary. This must be unflattened beforehand.
    Returns:
        None: The style dictionary is modified in place.

    NOTE: This function exists by design, not accident. It is useful for purposeful
    (e.g. node vs vertex) or historical (e.g. edge_padding vs edge_shrink) reasons.
    Either way, it softens the user experience without complicating the API.
    """

    # Accept "node" style as "vertex" style for user flexibility
    if "node" in style:
        style_node = style.pop("node")
        if "vertex" not in style:
            style["vertex"] = style_node
        else:
            # "node" style applies on TOP of "vertex" style
            update_style(style_node, style["vertex"])

    # NOTE: Deprecate edge_padding for edge_shrink
    for edgekey in ["edge", "leafedge"]:
        if "padding" in style.get(edgekey, {}):
            # shrink takes over
            if "shrink" in style[edgekey]:
                del style[edgekey]["padding"]
            else:
                style[edgekey]["shrink"] = style[edgekey].pop("padding")



"""
Make banner image for the package.
"""

import numpy as np
import pathlib
from matplotlib import patches
import matplotlib.pyplot as plt


def squareify(image):
    height, width, ncols = image.shape
    dmax = max(height, width)
    new = np.zeros((dmax, dmax, ncols), dtype=image.dtype)
    if height > width:
        diff = (height - width) // 2
        new[:, diff : diff + width, :] = image
        new[:, : diff + 3, :] = 1
        new[:, diff + width - 3 :] = 1
    elif width > height:
        diff = (width - height) // 2
        new[diff : diff + height, :, :] = image
        new[: diff + 3, :, :] = 1
        new[diff + height - 3 :, :, :] = 1

    zoom = 250
    diff = (new.shape[0] - zoom) // 2
    new = new[diff : diff + zoom, diff : diff + zoom, :]

    return new


if __name__ == "__main__":
    # Select random ones to fill a grid
    grid_shape = (4, 6)
    nimages = grid_shape[0] * grid_shape[1]

    # Folder with the images to patchwork
    image_folder_path = pathlib.Path(__file__).parent.parent / "docs" / "build" / "_images"
    images_all = list(image_folder_path.glob("*_thumb.png"))
    images = [images_all[i] for i in np.random.choice(len(images_all), nimages, replace=False)]

    k = 2.0
    kx, ky = 1.13, 0.98
    fig, ax = plt.subplots(figsize=(k * grid_shape[1], k * grid_shape[0]))
    for i in range(grid_shape[0]):
        for j in range(grid_shape[1]):
            ij = grid_shape[1] * i + j
            print(ij)
            img = plt.imread(images[ij])
            xmin = kx * (j + 0.5 * int(i % 2))
            xmax = xmin + kx * 1
            ymin = ky * (grid_shape[0] - 1 - i)
            ymax = ymin + ky * 1
            center = xmin + 0.5, ymin + 0.5
            patch = patches.RegularPolygon(
                center,
                6,
                radius=0.65,
                transform=ax.transData,
                facecolor="none",
                edgecolor="black",
                linewidth=1.5,
            )
            ax.add_patch(patch)
            # ax.text(
            #    center[0], center[1], str(ij), ha="center", va="center", fontsize=8, color="black"
            # )
            # Extend to square
            img = squareify(img)
            im = ax.imshow(img, extent=(xmin, xmax, ymin, ymax), aspect=1.0)
            im.set_clip_path(patch)
    ax.set(
        xlim=(-0.2 * kx, kx * (grid_shape[1] + 0.5 * int(grid_shape[0] > 1))),
        ylim=(-0.2 * ky, ky * (grid_shape[0] + 0.2)),
        aspect=1.0,
    )
    ax.set_axis_off()
    plt.ion()
    plt.show()

    fig.savefig(image_folder_path.parent.parent / "source" / "_static" / "banner.png", dpi=100)



from typing import (
    Any,
    Optional,
    Sequence,
    Iterable,
    Self,
)

from ...typing import (
    TreeDataProvider,
)


class SimpleTree:
    """A simple tree class for educational purposes.

    Properties:
        children: Children SimpleTree objects.
        branch_length: Length of the branch leading to this node/tree.
    """

    children: Sequence[Self] = []
    branch_length: float = 1
    name: str = ""

    @classmethod
    def from_dict(cls, data: dict) -> Self:
        """Create a SimpleTree from a dictionary.

        Parameters:
            data: A dictionary representation of the tree, with "children" as a list offset_transform
                child nodes and an optional "branch_length" property (float).

        Returns:
            An instance of SimpleTree constructed from the provided dictionary.
        """
        tree = cls()
        tree.branch_length = data.get("branch_length", 1)
        tree.name = data.get("name", "")
        tree.children = [cls.from_dict(child) for child in data.get("children", [])]
        return tree


class SimpleTreeDataProvider(TreeDataProvider):
    def is_rooted(self) -> bool:
        return True

    def get_root(self) -> Any:
        """Get the root node of the tree."""
        return self.tree

    def preorder(self) -> Iterable[dict[dict | str, Any]]:
        def _recur(node):
            yield node
            for child in node.children:
                yield from _recur(child)

        yield from _recur(self.tree)

    def postorder(self) -> Iterable[dict[dict | str, Any]]:
        def _recur(node):
            for child in node.children:
                yield from _recur(child)
            yield node

        yield from _recur(self.tree)

    def get_leaves(self) -> Sequence[Any]:
        def _recur(node):
            if len(node.children) == 0:
                yield node
            else:
                for child in node.children:
                    yield from _recur(child)

        return list(_recur(self.tree))

    @staticmethod
    def get_children(node: Any) -> Sequence[Any]:
        return node.children

    @staticmethod
    def get_branch_length(node: Any) -> Optional[float]:
        return node.branch_length

    @staticmethod
    def check_dependencies() -> bool:
        return True

    @staticmethod
    def tree_type():
        return SimpleTree

    def get_support(self):
        """Get support/confidence values for all nodes."""
        support_dict = {}
        for node in self.preorder():
            support_dict[node] = None
        return support_dict



"""
Tree element showroom
=====================

This example showcases plot elements for tree styling in ``iplotx``.
"""

import matplotlib.pyplot as plt
import iplotx as ipx

tree = {
    "name": "Vertex\nlabel",
    "children": [
        {
            "children": [
                {},
                {
                    "children": [{}, {}]
                },
            ],
        },
        {},
    ],
}
tree = ipx.ingest.providers.tree.simple.SimpleTree.from_dict(tree)

fig, ax = plt.subplots(figsize=(3.9, 3.3))
ipx.tree(
    tree,
    ax=ax,
    vertex_size=[0] * 6 + [(45, 35)],
    vertex_marker="r",
    vertex_facecolor=["none"] * 6 + ["orange"],
    vertex_alpha=0.5,
    vertex_edgecolor="black",
    vertex_linewidth=1.5,
    vertex_labels=True,
    vertex_label_hmargin=0,
    leaf_labels={tree.children[0].children[0]: "Leaf\nlabel"},
    leaf_label_bbox_facecolor="lightcoral",
    leaf_label_bbox_edgecolor="black",
    leaf_label_bbox_alpha=0.9,
    leaf_deep=True,
    cascade_facecolor={
        tree.children[0].children[1]: "gold",
    },
    edge_linewidth=3,
    edge_split_linewidth=3,
    edge_split_linestyle=["-."] + ["-"] * 6,
    edge_split_color=["tomato"] + ["black"] * 6,
    edge_capstyle="round",
    edge_split_capstyle="round",
    margins=(0.21, 0.08),
)
plt.ion(); plt.show()



"""
Module containing code to manipulate arrow visualisations in 3D, especially the EdgeArrow3DCollection class.
"""

from typing import (
    Sequence,
)
from math import atan2, cos, sin
import numpy as np
from matplotlib import (
    cbook,
)
from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.mplot3d.art3d import (
    Path3DCollection,
)

from ...utils.matplotlib import (
    _forwarder,
)
from ...edge.arrow import (
    EdgeArrowCollection,
)


@_forwarder(
    (
        "set_clip_path",
        "set_clip_box",
        "set_snap",
        "set_sketch_params",
        "set_animated",
        "set_picker",
    )
)
class EdgeArrow3DCollection(EdgeArrowCollection, Path3DCollection):
    """Collection of vertex patches for plotting."""

    def _update_before_draw(self) -> None:
        """Update the collection before drawing."""
        if (
            isinstance(self.axes, Axes3D)
            and hasattr(self, "do_3d_projection")
            and (self.axes.M is not None)
        ):
            self.do_3d_projection()

        # The original EdgeArrowCollection method for
        # _update_before_draw cannot be used because it
        # relies on paths, whereas edges are now a
        # Line3DCollection which uses segments.
        self.set_sizes(self._sizes, self.get_figure(root=True).dpi)

        if (not hasattr(self, "_z_markers_idx")) or (
            not isinstance(self._z_markers_idx, np.ndarray)
        ):
            return

        trans = self.get_offset_transform().transform

        # The do_3d_projection method above reorders the
        # arrow offsets in some way, so we might have to figure out
        # what edge index corres
        for i, ie in enumerate(self._z_markers_idx):
            segments_2d = self._edge_collection.get_segments()[ie]

            # We could reset the 3d projection here, might be a way to
            # skip the function call above.
            v2 = trans(segments_2d[-1])
            v1 = trans(segments_2d[-2])
            dv = v2 - v1
            theta = atan2(*(dv[::-1]))
            theta_old = self._angles[i]
            dtheta = theta - theta_old
            mrot = np.array([[cos(dtheta), sin(dtheta)], [-sin(dtheta), cos(dtheta)]])

            apath = self._paths[i]
            apath.vertices = apath.vertices @ mrot
            self._angles[i] = theta

    def draw(self, renderer) -> None:
        """Draw the collection of vertices in 3D.

        Parameters:
            renderer: The renderer to use for drawing.
        """
        with self._use_zordered_offset():
            with cbook._setattr_cm(self, _in_draw=True):
                EdgeArrowCollection.draw(self, renderer)


def arrow_collection_2d_to_3d(
    col: EdgeArrowCollection,
    zs: np.ndarray | float | Sequence[float] = 0,
    zdir: str = "z",
    depthshade: bool = True,
    axlim_clip: bool = False,
):
    """Convert a 2D EdgeArrowCollection to a 3D EdgeArrow3DCollection.

    Parameters:
        col: The 2D EdgeArrowCollection to convert.
        zs: The z coordinate(s) to use for the 3D vertices.
        zdir: The axis to use as the z axis (default is "z").
        depthshade: Whether to apply depth shading (default is True).
        axlim_clip: Whether to clip the vertices to the axes limits (default is False).
    """
    if not isinstance(col, EdgeArrowCollection):
        raise TypeError("vertices must be a EdgeArrowCollection")

    col.__class__ = EdgeArrow3DCollection
    col._offset_zordered = None
    col._depthshade = depthshade
    col._in_draw = False
    col.set_3d_properties(zs, zdir, axlim_clip)
